from typing import Mapping, Optional, Sequence, Tuple

import numpy as np

from starfish.core.types import ArrayLike, Axes, Coordinates, Number
from ..binary_mask import BinaryMaskCollection


def label_array_2d() -> Tuple[np.ndarray, Mapping[Coordinates, ArrayLike[Number]]]:
    """Convenience method to return a 2D label array with corresponding physical coordinates."""
    label_array = np.zeros((5, 6), dtype=np.int32)
    label_array[0] = 1
    label_array[3:5, 3:6] = 2
    label_array[-1, -1] = 0

    physical_ticks = {
        Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
        Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12, 13.2]}

    return label_array, physical_ticks


def binary_arrays_2d() -> Tuple[Sequence[np.ndarray], Mapping[Coordinates, ArrayLike[Number]]]:
    """Convenience method to return a set of 2D binary arrays with corresponding physical
    coordinates."""
    binary_arrays = [
        np.zeros((5, 6), dtype=np.bool),
        np.zeros((5, 6), dtype=np.bool),
    ]
    binary_arrays[0][0] = True
    binary_arrays[1][3:5, 3:6] = True
    binary_arrays[1][-1, -1] = False

    physical_ticks = {
        Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
        Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12, 13.2]}

    return binary_arrays, physical_ticks


def label_array_3d() -> Tuple[np.ndarray, Mapping[Coordinates, ArrayLike[Number]]]:
    """Convenience method to return a 3D label array with corresponding physical coordinates."""
    label_array = np.zeros((2, 5, 6), dtype=np.int32)
    label_array[0, 0] = 1
    label_array[:, 3:5, 3:6] = 2
    label_array[-1, -1, -1] = 0

    physical_ticks = {
        Coordinates.Z: [0.0, 1.0],
        Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
        Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12, 13.2]
    }

    return label_array, physical_ticks


def binary_arrays_3d() -> Tuple[Sequence[np.ndarray], Mapping[Coordinates, ArrayLike[Number]]]:
    """Convenience method to return a set of 2D binary arrays with corresponding physical
    coordinates."""
    binary_arrays = [
        np.zeros((2, 5, 6), dtype=np.bool),
        np.zeros((2, 5, 6), dtype=np.bool),
    ]
    binary_arrays[0][0, 0] = True
    binary_arrays[1][:, 3:5, 3:6] = True
    binary_arrays[1][-1, -1, -1] = False

    physical_ticks = {
        Coordinates.Z: [0.0, 1.0],
        Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
        Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12, 13.2]
    }

    return binary_arrays, physical_ticks


def binary_mask_collection_2d(
        label_array: Optional[np.ndarray] = None,
        physical_ticks: Optional[Mapping[Coordinates, ArrayLike[Number]]] = None,
        pixel_ticks: Optional[Mapping[Axes, ArrayLike[int]]] = None):
    """Convenience method to return a 2D binary mask collection."""
    if label_array is None or physical_ticks is None:
        new_label_array, new_physical_ticks = label_array_2d()
        label_array = label_array or new_label_array
        physical_ticks = physical_ticks or new_physical_ticks

    return BinaryMaskCollection.from_label_array_and_ticks(
        label_array,
        pixel_ticks,
        physical_ticks,
        None
    )


def binary_mask_collection_3d(
        label_array: Optional[np.ndarray],
        physical_ticks: Optional[Mapping[Coordinates, ArrayLike[Number]]],
        pixel_ticks: Optional[Mapping[Axes, ArrayLike[int]]]):
    """Convenience method to return a 3D binary mask collection."""
    if label_array is None or physical_ticks is None:
        new_label_array, new_physical_ticks = label_array_3d()
        label_array = label_array or new_label_array
        physical_ticks = physical_ticks or new_physical_ticks

    return BinaryMaskCollection.from_label_array_and_ticks(
        label_array,
        pixel_ticks,
        physical_ticks,
        None
    )



import itertools
import multiprocessing
import os
from pathlib import Path
from typing import cast, Mapping, Optional, Sequence

from slicedimage import ImageFormat

from starfish.core.experiment.builder import (
    tile_fetcher_factory,
    TileFetcher,
    TileIdentifier,
)
from starfish.core.experiment.builder.structured_formatter import format_structured_dataset
from starfish.core.experiment.builder.test.structured_formatter.util import (
    render_coordinates_to_rows,
    UniqueTiles,
    write_coordinates_csv,
    write_tile_data,
)
from starfish.core.experiment.experiment import Experiment, FieldOfView
from starfish.core.image import Filter
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Coordinates, CoordinateValue


def test_multiprocessing_workflow(tmpdir):
    exp = build_experiment_with_multiple_unaligned_fovs(Path(tmpdir))
    assert len(exp.fovs()) > 1
    for fov in exp.fovs():
        assert len(list(fov.get_images(FieldOfView.PRIMARY_IMAGES))) > 1

    all_imagestacks = itertools.chain.from_iterable(
        fov.get_images(FieldOfView.PRIMARY_IMAGES)
        for fov in exp.fovs())

    with multiprocessing.Pool(processes=2) as pool:
        output_imagestacks = pool.map(filter_fn, all_imagestacks)

    for output_imagestack in output_imagestacks:
        assert isinstance(output_imagestack, ImageStack)


def filter_fn(imagestack: ImageStack) -> Optional[ImageStack]:
    f = Filter.Clip()
    return f.run(imagestack)


def build_experiment_with_multiple_unaligned_fovs(
        tmpdir_path: Path,
        fovs=(0, 1, 2, 3),
        rounds=(1, 2, 4),
        chs=(2, 3, 4),
        zplanes=(0, 1, 2),
        tile_height=100,
        tile_width=60,
        fov_to_x_coords={
            0: (0.0, 0.1),
            1: (0.0, 0.1),
            2: (0.1, 0.2),
            3: (0.1, 0.2),
        },
        fov_to_y_coords={
            0: (0.0, 0.1),
            1: (0.1, 0.2),
            2: (0.0, 0.1),
            3: (0.1, 0.2),
        },
        zplane_to_coords={0: 0.20, 1: 0.25, 2: 0.3},
        # default values are mutable, but it's for readability reasons!
) -> Experiment:
    """Write the tiles for a multi-fov unaligned (physical coordinates) regular (the dimensions have
    the same cardinality) image.  Then build an experiment from the tiles.  Finally, load the
    resulting experiment and return it."""
    tile_identifiers: Sequence[TileIdentifier] = [
        TileIdentifier(fov_id, round_label, ch_label, zplane_label)
        for fov_id in fovs
        for round_label in rounds
        for ch_label in chs
        for zplane_label in zplanes
    ]
    tile_fetcher: TileFetcher = tile_fetcher_factory(
        UniqueTiles,
        pass_tile_indices=True,
        fovs=fovs,
        rounds=rounds,
        chs=chs,
        zplanes=zplanes,
        tile_height=tile_height,
        tile_width=tile_width,
    )
    tile_coordinates: Mapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]] = {
        tile_identifier: {
            Coordinates.X: cast(CoordinateValue, tuple(
                coordinate + float(tile_identifier.round_label) * 0.0001
                for coordinate in fov_to_x_coords[tile_identifier.fov_id])),
            Coordinates.Y: fov_to_y_coords[tile_identifier.fov_id],
            Coordinates.Z: zplane_to_coords[tile_identifier.zplane_label],
        }
        for tile_identifier in tile_identifiers
    }

    write_tile_data(
        tmpdir_path,
        FieldOfView.PRIMARY_IMAGES,
        ImageFormat.TIFF,
        tile_identifiers,
        tile_fetcher)

    coordinates_csv_path = tmpdir_path / "coordinates.csv"
    rows = render_coordinates_to_rows(tile_coordinates)
    write_coordinates_csv(coordinates_csv_path, rows, True)

    format_structured_dataset(
        os.fspath(tmpdir_path),
        os.fspath(coordinates_csv_path),
        os.fspath(tmpdir_path),
        ImageFormat.TIFF,
        False,
    )

    # load the data and verify it.
    return Experiment.from_json(os.fspath(tmpdir_path / "experiment.json"))



"""
Tests for codebook.per_round_max_decode method
"""

import numpy as np
import pandas as pd
import pytest

from starfish import IntensityTable
from starfish.core.types import Axes, Features, SpotAttributes
from ..codebook import Codebook


def intensity_table_factory(data: np.ndarray=np.array([[[0, 3], [4, 0]]])) -> IntensityTable:
    """IntensityTable with a single feature that was measured over 2 channels and 2 rounds."""

    # generates spot attributes equal in size to the number of passed features.
    # each attribute has coordinates (z, y, x) equal to the feature index, and radius 1.
    spot_attributes_data = pd.DataFrame(
        data=np.array([[i, i, i, 1] for i in np.arange(data.shape[0])]),
        columns=[Axes.ZPLANE, Axes.Y, Axes.X, Features.SPOT_RADIUS]
    )

    spot_attributes = SpotAttributes(spot_attributes_data)
    intensity_table = IntensityTable.from_spot_data(
        data, spot_attributes,
        round_values=np.arange(data.shape[1]),
        ch_values=np.arange(data.shape[2]),
    )
    return intensity_table


def codebook_factory() -> Codebook:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    codebook_array = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: 'GENE_A'
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: 'GENE_B'
        },
    ]
    return Codebook.from_code_array(codebook_array)


def test_intensity_tables_with_different_numbers_of_codes_or_channels_throw_value_error():
    """
    The test passes a 3-round and 1-round IntensityTable to a 2-round codebook. Both should
    raise a ValueError.
    """
    data = np.array(
        [[[4, 4],
          [3, 0],
          [1, 2],
          ]
         ]
    )
    codebook = codebook_factory()
    intensities = intensity_table_factory(data)
    with pytest.raises(ValueError):
        codebook.decode_per_round_max(intensities)

    with pytest.raises(ValueError):
        codebook.decode_per_round_max(intensities.sel(round=1))


def test_example_intensity_decodes_to_gene_a():
    """The single intensity should decode to gene a. Verify that it does."""
    intensities = intensity_table_factory()
    codebook = codebook_factory()

    decoded_intensities = codebook.decode_per_round_max(intensities)

    assert decoded_intensities[Features.TARGET] == 'GENE_A'
    assert np.array_equal(decoded_intensities[Features.PASSES_THRESHOLDS], [True])
    assert np.array_equal(decoded_intensities[Features.DISTANCE], [0])


def test_example_intensity_that_has_no_match_is_assigned_nan():
    """
    The decoder assigns nan when it doesn't receive a match, but the nan is converted to a
    string due to the object dtype
    """
    no_match = np.array([[[3, 0], [0, 4]]])
    intensities = intensity_table_factory(no_match)
    codebook = codebook_factory()

    decoded_intensities = codebook.decode_per_round_max(intensities)

    # even though the feature did not decode, it should still be returned
    assert decoded_intensities.sizes[Features.AXIS] == 1

    # assert that the feature did not decode
    assert decoded_intensities[Features.TARGET] == 'nan'
    assert np.array_equal(decoded_intensities[Features.PASSES_THRESHOLDS], [False])

    # distance is calculated as the fraction of signal NOT in the selected channel. Here all of
    # the signal is in the max channel, so distance is 0.
    # Admittedly, this is a little bit confusing, since distance here measures both how correct
    # a code is, but also how wrong one is when it _doesn't_ decode properly.
    assert np.array_equal(decoded_intensities[Features.DISTANCE], [0])


def test_argmax_selects_the_last_equal_intensity_channel_and_decodes_consistently():
    """
    When two channels are tied, argmax selects the last first one. This is the code for Gene A:

    [[0, 1],
     [1, 0]]

    both of these observations could potentially match the code for GENE_A:

    [[4, 3],
     [4, 0]]

    and

    [[0, 3],
     [4, 3]]

    The first is interpreted as argmax as:

    [[1, 1],
     [0, 0]]

    which fails to match, while the second is interpreted as

    [[0, 1],
     [1, 0]]

    and matches GENE_A.

    This test also verifies that the decoder can operate on multiple codes.
    """

    data = np.array(
        [[[4, 3],  # this code is decoded "wrong"
          [4, 0]],
         [[0, 3],  # this code is decoded "right"
          [4, 3]]]
    )
    intensities = intensity_table_factory(data)
    codebook = codebook_factory()

    decoded_intensities = codebook.decode_per_round_max(intensities)
    assert np.array_equal(decoded_intensities[Features.TARGET].values, ['nan', 'GENE_A'])


def test_argmax_does_not_select_first_code():
    """
    When all the channels in a round are uniform, argmax erroneously picks the first channel as the
    max.  In this case, it incorrectly assigns the wrong code for that round.  This test ensures
    that the workaround we put in for this works correctly.
    """

    data = np.array(
        [[[0.0, 1.0],
          [1.0, 1.0]],  # this round is uniform, so it will erroneously be decoded as the first ch.
         [[0.0, 1.0],
          [1.0, 0.0]]]
    )
    intensities = intensity_table_factory(data)
    codebook = codebook_factory()

    decoded_intensities = codebook.decode_per_round_max(intensities)
    assert np.array_equal(decoded_intensities[Features.TARGET].values, ['nan', 'GENE_A'])


def test_feature_round_all_nan():
    """
    When all the channels in a round are NaN, argmax chokes.  This test ensures that the workaround
    we put in for this works correctly.
    """

    data = np.array(
        [[[0.0, 1.0],
          [np.nan, np.nan]],
         [[0.0, 1.0],
          [1.0, 0.0]]]
    )
    intensities = intensity_table_factory(data)
    codebook = codebook_factory()

    decoded_intensities = codebook.decode_per_round_max(intensities)
    assert np.array_equal(decoded_intensities[Features.TARGET].values, ['nan', 'GENE_A'])



import numpy as np
import xarray as xr

from starfish import DecodedIntensityTable, IntensityTable
from starfish.core.codebook.test.factories import codebook_array_factory, loaded_codebook
from starfish.core.types import Coordinates, Features
from ..overlap import Area


def synthetic_intensity_table() -> IntensityTable:
    return IntensityTable.synthetic_intensities(loaded_codebook(), n_spots=2)


def synthetic_decoded_intensity_table(
        codebook,
        num_z: int = 12,
        height: int = 50,
        width: int = 40,
        n_spots: int = 10,
        mean_fluor_per_spot: int = 200,
        mean_photons_per_fluor: int = 50,
) -> DecodedIntensityTable:
    """
    Creates an IntensityTable with synthetic spots, that correspond to valid
    codes in a provided codebook.

    Parameters
    ----------
    codebook : Codebook
        Starfish codebook object.
    num_z : int
        Number of z-planes to use when localizing spots.
    height : int
        y dimension of each synthetic plane.
    width : int
        x dimension of each synthetic plane.
    n_spots : int
        Number of spots to generate.
    mean_fluor_per_spot : int
         Mean number of fluorophores per spot.
    mean_photons_per_fluor : int
        Mean number of photons per fluorophore.

    Returns
    -------
    DecodedIntensityTable
    """

    intensities = IntensityTable.synthetic_intensities(
        codebook,
        num_z=num_z,
        height=height,
        width=width,
        n_spots=n_spots,
        mean_fluor_per_spot=mean_fluor_per_spot,
        mean_photons_per_fluor=mean_photons_per_fluor
    )
    targets = np.random.choice(
        codebook.coords[Features.TARGET], size=n_spots, replace=True)

    return DecodedIntensityTable.from_intensity_table(intensities, targets=(Features.AXIS, targets))


def assign_synthetic_targets(intensities: IntensityTable) -> DecodedIntensityTable:
    """
    Assign fake target values to the given IntensityTable

    Parameters
    ----------
    intensities : IntensityTable
        intensity_table to assign target values to

    Returns
    -------
    DecodedIntensityTable
    """
    intensities = DecodedIntensityTable(intensities)
    return DecodedIntensityTable.from_intensity_table(
        intensities,
        targets=(Features.AXIS, np.random.choice(list('ABCD'), size=20)),
        distances=(Features.AXIS, np.random.rand(20)))


def create_intensity_table_with_coords(area: Area, n_spots: int=10) -> IntensityTable:
    """
    Creates a 50X50 intensity table with physical coordinates within
    the given Area.

    Parameters
    ----------
    area: Area
        The area of physical space the IntensityTable should be defined over
    n_spots:
        Number of spots to add to the IntensityTable
    """
    codebook = codebook_array_factory()
    it = IntensityTable.synthetic_intensities(
        codebook,
        num_z=1,
        height=50,
        width=50,
        n_spots=n_spots
    )
    # intensity table 1 has 10 spots, xmin = 0, ymin = 0, xmax = 2, ymax = 1
    it[Coordinates.X.value] = xr.DataArray(np.linspace(area.min_x, area.max_x, n_spots),
                                           dims=Features.AXIS)
    it[Coordinates.Y.value] = xr.DataArray(np.linspace(area.min_y, area.max_y, n_spots),
                                           dims=Features.AXIS)
    return it



import os

from starfish.types import Axes, Features
from .factories import simple_codebook_array
from ..codebook import Codebook


def test_to_json(tmp_path):
    code_array = simple_codebook_array()
    codebook = Codebook.from_code_array(code_array)
    codebook_path = tmp_path / "codebook.json"
    codebook.to_json(codebook_path)

    loaded_codebook = Codebook.open_json(os.fspath(codebook_path))
    assert codebook.equals(loaded_codebook)


def test_to_json_multiple_codes_for_target(tmp_path):
    code_array = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "SCUBE2"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "BRCA"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "SCUBE2"
        }
    ]
    codebook = Codebook.from_code_array(code_array)
    codebook_path = tmp_path / "codebook.json"
    codebook.to_json(codebook_path)

    loaded_codebook = Codebook.open_json(os.fspath(codebook_path))
    assert codebook.equals(loaded_codebook)



from collections import OrderedDict

import numpy as np

from starfish import data
from starfish.core.types import Axes, PhysicalCoordinateTypes
from .factories import imagestack_with_coords_factory
from .imagestack_test_utils import verify_physical_coordinates
from ..imagestack import ImageStack


def test_max_projection_preserves_dtype():
    original_dtype = np.float32
    array = np.ones((2, 2, 2), dtype=original_dtype)
    image = ImageStack.from_numpy(array.reshape((1, 1, 2, 2, 2)))

    max_projection = image.reduce((Axes.CH, Axes.ROUND, Axes.ZPLANE), "max")
    assert max_projection.xarray.dtype == original_dtype


X_COORDS = 1, 2
Y_COORDS = 4, 6
Z_COORDS = 1, 3


def test_max_projection_preserves_coordinates():
    e = data.ISS(use_test_data=True)
    nuclei = e.fov().get_image('nuclei')
    nuclei_proj = nuclei.reduce((Axes.CH, Axes.ROUND, Axes.ZPLANE), "max")
    # Since this data already has only 1 round, 1 ch, 1 zplane
    # let's just assert that the max_proj operation didn't change anything
    assert nuclei.xarray.equals(nuclei_proj.xarray)

    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 3), (Axes.Y, 10), (Axes.X, 10)])

    # Create stack with coordinates, verify coords unaffected by max_poj
    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, X_COORDS[0]),
                                   (PhysicalCoordinateTypes.X_MAX, X_COORDS[1]),
                                   (PhysicalCoordinateTypes.Y_MIN, Y_COORDS[0]),
                                   (PhysicalCoordinateTypes.Y_MAX, Y_COORDS[1]),
                                   (PhysicalCoordinateTypes.Z_MIN, Z_COORDS[0]),
                                   (PhysicalCoordinateTypes.Z_MAX, Z_COORDS[1])])

    stack = imagestack_with_coords_factory(stack_shape, physical_coords)

    stack_proj = stack.reduce((Axes.CH, Axes.ROUND, Axes.ZPLANE), "max")
    expected_z = np.average(Z_COORDS)
    verify_physical_coordinates(stack_proj, X_COORDS, Y_COORDS, expected_z)



import warnings
from typing import (
    Iterable,
    MutableMapping,
    Optional,
    Union
)

import numpy as np

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import (
    ArrayLike,
    Axes,
    Clip,
    Coordinates,
    FunctionSource,
    FunctionSourceBundle,
    Levels,
    Number,
)
from starfish.core.util.levels import levels
from ._base import FilterAlgorithm


class Reduce(FilterAlgorithm):
    """
    Reduces the cardinality of one or more axes to 1 by applying a function across those axes.

    Parameters
    ----------
    dims : Iterable[Union[Axes, str]]
        one or more Axes to reduce over
    func : Union[str, FunctionSourceBundle]
        Function to apply across the dimension(s) specified by ``dims``.

        If this value is a string, then the ``module`` parameter is consulted to determine which
        python package is used to find the function.  If ``module`` is not specified, then the
        default is :py:attr:`FunctionSource.np`.

        If this value is a ``FunctionSourceBundle``, then the python package and module name is
        obtained from the bundle.

        Some common examples for the np FunctionSource:

        - amax: maximum intensity projection (applies np.amax)
        - max: maximum intensity projection (this is an alias for amax and applies np.amax)
        - mean: take the mean across the dim(s) (applies np.mean)
        - sum: sum across the dim(s) (applies np.sum)
    module : Optional[FunctionSource]
        Python module that serves as the source of the function.  It must be listed as one of the
        members of :py:class:`FunctionSource`.

        Currently, the supported FunctionSources are:
        - ``np``: the top-level package of numpy
        - ``scipy``: the top-level package of scipy

        This is being deprecated in favor of specifying the function as a ``FunctionSourceBundle``.
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.

    Examples
    --------
    Reducing via max projection.
        >>> from starfish.core.imagestack.test.factories.synthetic_stack import synthetic_stack
        >>> from starfish.image import Filter
        >>> from starfish.types import Axes
        >>> stack = synthetic_stack()
        >>> reducer = Filter.Reduce({Axes.ROUND}, func="max")
        >>> max_proj = reducer.run(stack)

    Reducing via linalg.norm
        >>> from starfish.core.imagestack.test.factories.synthetic_stack import synthetic_stack
        >>> from starfish.image import Filter
        >>> from starfish.types import Axes, FunctionSource
        >>> stack = synthetic_stack()
        >>> reducer = Filter.Reduce(
                {Axes.ROUND},
                func=FunctionSource.scipy("linalg.norm"),
                ord=2,
            )
        >>> norm = reducer.run(stack)

    See Also
    --------
    starfish.core.types.Axes

    """

    def __init__(
        self,
            dims: Iterable[Union[Axes, str]],
            func: Union[str, FunctionSourceBundle] = "max",
            module: Optional[FunctionSource] = None,
            clip_method: Optional[Clip] = None,
            level_method: Optional[Levels] = None,
            **kwargs
    ) -> None:
        self.dims: Iterable[Axes] = set(Axes(dim) for dim in dims)
        if isinstance(func, str):
            if module is not None:
                warnings.warn(
                    f"The module parameter is being deprecated.  Use "
                    f"`func=FunctionSource.{module.name}{func} instead.",
                    DeprecationWarning)
            else:
                module = FunctionSource.np
            self.func = module(func)
        elif isinstance(func, FunctionSourceBundle):
            if module is not None:
                raise ValueError(
                    f"When passing in the function as a `FunctionSourceBundle`, module should not "
                    f"be set."
                )
            self.func = func
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)
        self.kwargs = kwargs

    _DEFAULT_TESTING_PARAMETERS = {"dims": ['r'], "func": 'max'}

    def run(
            self,
            stack: ImageStack,
            *args,
    ) -> ImageStack:
        """Performs the dimension reduction with the specifed function

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.

        Returns
        -------
        ImageStack :
            Return the results of filter as a new stack.

        """

        # Apply the reducing function
        reduced = stack.xarray.reduce(
            self.func.resolve(), dim=[dim.value for dim in self.dims], **self.kwargs)

        # Add the reduced dims back and align with the original stack
        reduced = reduced.expand_dims(tuple(dim.value for dim in self.dims))
        reduced = reduced.transpose(*stack.xarray.dims)

        if self.level_method == Levels.CLIP:
            reduced = levels(reduced)
        elif self.level_method in (Levels.SCALE_BY_CHUNK, Levels.SCALE_BY_IMAGE):
            reduced = levels(reduced, rescale=True)
        elif self.level_method in (
                Levels.SCALE_SATURATED_BY_CHUNK, Levels.SCALE_SATURATED_BY_IMAGE):
            reduced = levels(reduced, rescale_saturated=True)

        # Update the physical coordinates
        physical_coords: MutableMapping[Coordinates, ArrayLike[Number]] = {}
        for axis, coord in (
                (Axes.X, Coordinates.X),
                (Axes.Y, Coordinates.Y),
                (Axes.ZPLANE, Coordinates.Z)):
            if axis in self.dims:
                # this axis was projected out of existence.
                assert coord.value not in reduced.coords
                physical_coords[coord] = [np.average(stack._data.coords[coord.value])]
            else:
                physical_coords[coord] = reduced.coords[coord.value]
        reduced_stack = ImageStack.from_numpy(reduced.values, coordinates=physical_coords)

        return reduced_stack



import os
import tempfile
from typing import Optional

from starfish.core.types import Axes, Features
from ..codebook import Codebook


def simple_codebook_array():
    return [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "SCUBE2"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "BRCA"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 1, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 0, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "ACTB"
        }
    ]


def loaded_codebook():
    tfn: Optional[str] = None
    try:
        with tempfile.NamedTemporaryFile(delete=False) as tf:
            tfn = tf.name
            codebook = Codebook.from_code_array(simple_codebook_array())
            codebook.to_json(tfn)

        result = Codebook.open_json(tf.name, n_channel=2, n_round=2)
    finally:
        if tfn is not None:
            os.remove(tfn)

    return result


def codebook_array_factory() -> Codebook:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    data = [
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 0, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Axes.ROUND.value: 0, Axes.CH.value: 2, Features.CODE_VALUE: 1},
                {Axes.ROUND.value: 1, Axes.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    return Codebook.from_code_array(data)



import copy
import json
import posixpath
import sys
import warnings
from pathlib import Path
from typing import Any, Dict, IO, Iterator, List, Optional, Union

from jsonschema import Draft4Validator, RefResolver, ValidationError
from pkg_resources import resource_filename
from semantic_version import Version
from slicedimage import VERSIONS as SLICEDIMAGE_VERSIONS

from starfish.core.codebook._format import CURRENT_VERSION as CODEBOOK_CURRENT_VERSION
from starfish.core.experiment.version import CURRENT_VERSION as EXPERIMENT_CURRENT_VERSION


TILESET_CURRENT_VERSION = Version(SLICEDIMAGE_VERSIONS[-1].VERSION)
package_name = 'starfish'


def _get_absolute_schema_path(schema_name: str) -> str:
    """turn the name of the schema into an absolute path by joining it to <package_root>/schema."""
    return resource_filename("starfish", posixpath.join("spacetx_format", "schema", schema_name))


class SpaceTxValidator:

    def __init__(self, schema: str) -> None:
        """create a validator for a json-schema compliant spaceTx specification file

        Parameters
        ----------
        schema : str
            file path to schema

        """
        self._schema: Dict = self.load_json(schema)
        self._validator: Draft4Validator = self._create_validator(self._schema)

    @staticmethod
    def _create_validator(schema: Dict) -> Draft4Validator:
        """resolve $ref links in a loaded json schema and return a validator

        Parameters
        ----------
        schema : Dict
            loaded json schema

        Returns
        -------
        Draft4Validator :
            json-schema validator specific to the supplied schema, with references resolved

        """

        # Note: we are using 5.0.0 here as the first known file. It does *not* need to
        # be upgraded with each version bump since only the dirname is used.
        experiment_schema_path = Path(resource_filename(
            package_name, "spacetx_format/schema/experiment_5.0.0.json"))

        package_root_path = experiment_schema_path.parent.parent
        base_uri = f"{package_root_path.as_uri()}/"
        resolver = RefResolver(base_uri, schema)
        return Draft4Validator(schema, resolver=resolver)

    @staticmethod
    def load_json(json_file: str) -> Dict:
        with open(json_file, 'rb') as f:
            return json.load(f)

    @staticmethod
    def _recurse_through_errors(error_iterator: Iterator[ValidationError],
                                level: int=0,
                                filename: str=None) -> None:
        """Recurse through ValidationErrors, printing message and schema path

        Parameters
        ----------
        error_iterator : Iterator[ValidationError]
            iterator over ValidationErrors that occur during validation
        level : int
            current level of recursion
        filename: str
            informational string regarding the source file of the given object

        """
        fmt = "\n{stars} {message}\n"
        fmt += "\tSchema:         \t{schema}\n"
        fmt += "\tSubschema level:\t{level}\n"
        fmt += "\tPath to error:  \t{path}\n"
        if filename:
            fmt += "\tFilename:       \t{filename}\n"
        for error in error_iterator:
            message = fmt.format(
                stars="***" * level, level=str(level),
                path="/".join([str(x)for x in error.absolute_schema_path]),
                message=error.message, cause=error.cause, schema=error.schema.get("$id", "unknown"),
                filename=filename,
            )
            warnings.warn(message)
            if error.context:
                level += 1
                SpaceTxValidator._recurse_through_errors(error.context, level=level)

    def validate_file(self, target_file: str) -> bool:
        """validate a target file, returning True if valid and False otherwise

        Parameters
        ----------
        target_file : str
            path or URL to a target json object to be validated against the schema passed to this
            object's constructor

        Returns
        -------
        bool :
            True, if object valid, else False

        Examples
        --------
        Validate a codebook file::

            >>> from pkg_resources import resource_filename
            >>> from starfish.core.spacetx_format.util import SpaceTxValidator
            >>> schema_path = resource_filename(
                    "starfish", "spacetx_format/schema/codebook/codebook.json")
            >>> validator = SpaceTxValidator(schema_path)
            >>> if not validator.validate_file(your_codebook_filename):
            >>>     raise Exception("invalid")

        """
        target_object = self.load_json(target_file)
        return self.validate_object(target_object, target_file)

    def validate_object(
            self,
            target_object: Union[dict, list],
            target_file: str=None,
    ) -> bool:
        """validate a loaded json object, returning True if valid, and False otherwise

        Parameters
        ----------
        target_object : Dict
            loaded json object to be validated against the schema passed to this object's
            constructor
        target_file : str
            informational string regarding the source file of the given object

        Returns
        -------
        bool :
            True, if object valid, else False

        Examples
        --------
        Validate an experiment json string ::

            >>> from pkg_resources import resource_filename
            >>> from starfish.core.spacetx_format.util import SpaceTxValidator
            >>> schema_path = resource_filename("starfish", "spacetx_format/schema/experiment.json")
            >>> validator = SpaceTxValidator(schema_path)
            >>> if not validator.validate_object(your_experiment_object):
            >>>     raise Exception("invalid")

        """

        if self._validator.is_valid(target_object):
            return True
        else:
            es: Iterator[ValidationError] = self._validator.iter_errors(target_object)
            self._recurse_through_errors(es, filename=target_file)
            return False

    def fuzz_object(
            self,
            target_object: Union[dict, list],
            target_file: str=None,
            out: IO=sys.stdout,
    ) -> None:
        """performs mutations on the given object and tests for validity.

        A representation of the validity is printed to the given output stream.

        Parameters
        ----------
        target_object : Dict
            loaded json object to be validated against the schema passed to this object's
            constructor
        target_file : str
            informational string regarding the source file of the given object
        out : IO
            output stream for printing

        """

        if target_file:
            out.write(f"> Fuzzing {target_file}...\n")
        else:
            out.write("> Fuzzing unknown...\n")
        fuzzer = Fuzzer(self._validator, target_object, out)
        fuzzer.fuzz()


class Fuzzer(object):

    def __init__(self, validator: Draft4Validator, obj: Any, out: IO=sys.stdout) -> None:
        """create a fuzzer which will check different situations against the validator

        Parameters
        ----------
        validator : SpaceTxValidator
            validator which should match the given object type
        obj : Any
            JSON-like object which will be checked against the validator
        out : IO
            output stream for printing

        """
        self.validator = validator
        self.obj = obj
        self.out = out
        self.stack: Optional[List[Any]] = None

    def fuzz(self) -> None:
        """prints to the out field the state of the object tree after types of fuzzing

        Each line is prefixed by the output of {state()} followed by a YAML-like
        representation of the branch of the object tree.
        """
        header = f"{self.state()}"
        header += "If the letter is present, mutation is valid!"
        self.out.write(f"{header}\n")
        self.out.write("".join([x in ("\t", "\n") and x or "-" for x in header]))
        self.out.write("\n")
        self.stack = []
        try:
            self.descend(self.obj)
        finally:
            self.stack = None

    def state(self) -> str:
        """primary driver for the checks of individual trees

        Returns
        -------
        str :
            space-separated representation of the fuzzing conditions.
            If a letter is present, then mutation leaves the tree in
            a valid state:

             A: inserting a fake key or appending to a list
             D: deleting a key or index
             I: converting value to an integer
             I: converting value to a string
             M: converting value to an empty dict
             L: converting value to an empty list

        """
        rv = [
            Add().check(self),
            Del().check(self),
            Change("I", lambda *args: 123456789).check(self),
            Change("S", lambda *args: "fake").check(self),
            Change("M", lambda *args: dict()).check(self),
            Change("L", lambda *args: list()).check(self),
        ]
        return ' '.join(rv) + "\t"

    def descend(self, obj: Any, depth: int=0, prefix: str="") -> None:
        """walk a JSON-like object tree printing the state of the tree
        at each level. A YAML representation is used for simplicity.

        Parameters
        ----------
        obj : Any
            JSON-like object tree
        depth : int
            depth in the tree that is currently being evaluated
        prefix : str
            value which should be prepended to printouts at this level
        """
        if self.stack is None:
            raise Exception("invalid state")
        if isinstance(obj, list):
            for i, o in enumerate(obj):
                depth += 1
                self.stack.append(i)
                self.descend(o, depth, prefix="- ")
                self.stack.pop()
                depth -= 1
        elif isinstance(obj, dict):
            for k in obj:
                # This is something of a workaround in that we need a special
                # case for object keys since no __getitem__ method will suffice.
                self.stack.append((k,))
                self.out.write(f"{self.state()}{' ' * depth}{prefix}{k}:\n")
                self.stack.pop()
                if prefix == "- ":
                    prefix = "  "
                depth += 1
                self.stack.append(k)
                self.descend(obj[k], depth, prefix="  " + prefix)
                self.stack.pop()
                depth -= 1
        else:
            self.out.write(f"{self.state()}{' ' * depth}{prefix}{obj}\n")

class Checker(object):

    @property
    def LETTER(self) -> str:
        return "?"

    def check(self, fuzz: Fuzzer) -> str:
        """create a copy of the current state of the object tree,
        mutate it, and run it through is_valid on the validator.

        Parameters
        ----------
        fuzz : Fuzzer
            the containing instance

        Returns
        -------
        str :
            A single character string representation of the check

        """
        # Don't mess with the top level
        if fuzz.stack is None:
            return self.LETTER
        if not fuzz.stack:
            return "-"
        # Operate on a copy for mutating
        dupe = copy.deepcopy(fuzz.obj)
        target = dupe
        for level in fuzz.stack[0:-1]:
            target = target.__getitem__(level)
        self.handle(fuzz, target)
        valid = fuzz.validator.is_valid(dupe)
        return valid and self.LETTER or "."

    def handle(self, fuzz, target):
        raise NotImplementedError()

class Add(Checker):

    @property
    def LETTER(self) -> str:
        return "A"

    def handle(self, fuzz, target):
        if isinstance(target, dict):
            target["fake"] = "!"
        elif isinstance(target, list):
            target.append("!")
        else:
            raise Exception("unknown")

class Del(Checker):

    @property
    def LETTER(self) -> str:
        return "D"

    def handle(self, fuzz, target):
        key = fuzz.stack[-1]
        if isinstance(key, tuple):
            key = key[0]
        target.__delitem__(key)

class Change(Checker):

    @property
    def LETTER(self) -> str:
        return self.letter

    def __init__(self, letter, call):
        self.letter = letter
        self.call = call

    def handle(self, fuzz, target):
        key = fuzz.stack[-1]
        if isinstance(key, tuple):
            key = key[0]
        target.__setitem__(key, self.call())


def get_schema_path(
        schema: str,
        doc: Optional[dict] = None,
        version: Optional[Version] = None,
) -> str:
    """lookup the absolute schema path, including version, based
    on the given parameters

    Parameters
    ----------
    schema : str
        A portion of the schema path. The heuristic applied is that if any
        of the strings "codebook", "experiment", "fov_manifest", "coordinates",
        "indices", "tiles", or "field_of_view" is found in the string, then the
        appropriate schema is returned. Otherwise an exception is raised.
    doc: dict
        If provided, the "version" key will be read from it.
    version: Version
        Default version string to use if not found in the doc.

    Returns
    -------
    str :
        absolute schema path, never None.

    """

    if doc is not None:
        version_str = doc.get("version", version)
        if version_str is not None:
            version = Version(version_str)

    if version is None:
        raise ValueError("Could not find the version of the schema to validate against")

    if "codebook" in schema:
        path = f"codebook_{version}/codebook.json"
    elif "experiment" in schema:
        path = f"experiment_{version}.json"
    elif "fov_manifest" in schema:
        path = f"fov_manifest_{version}.json"
    elif "coordinates" in schema:
        path = f"field_of_view_{version}/tiles/coordinates.json"
    elif "indices" in schema:
        path = f"field_of_view_{version}/tiles/indices.json"
    elif "tiles" in schema:
        path = f"field_of_view_{version}/tiles/tiles.json"
    elif "field_of_view" in schema:
        path = f"field_of_view_{version}/field_of_view.json"
    else:
        raise Exception(f"Unknown schema: {schema}")
    return _get_absolute_schema_path(path)


class CodebookValidator(SpaceTxValidator):
    """
    Subclass of SpaceTxValidator which enforces the use of the "codebook" schema
    as returned by get_schema_path.
    """

    def __init__(self, doc: Optional[dict] = None, version: Optional[Version] = None):
        super(CodebookValidator, self).__init__(get_schema_path("codebook", doc, version))


class ExperimentValidator(SpaceTxValidator):
    """
    Subclass of SpaceTxValidator which enforces the use of the "experiment" schema
    as returned by get_schema_path.
    """

    def __init__(self, doc: Optional[dict] = None, version: Optional[Version] = None):
        super(ExperimentValidator, self).__init__(get_schema_path("experiment", doc, version))


class FOVValidator(SpaceTxValidator):
    """
    Subclass of SpaceTxValidator which enforces the use of the "field_of_view" schema
    as returned by get_schema_path.
    """

    def __init__(self, doc: Optional[dict] = None, version: Optional[Version] = None):
        super(FOVValidator, self).__init__(get_schema_path("field_of_view", doc, version))


class ManifestValidator(SpaceTxValidator):
    """
    Subclass of SpaceTxValidator which enforces the use of the "fov_manifest" schema
    as returned by get_schema_path.
    """

    def __init__(self, doc: Optional[dict] = None, version: Optional[Version] = None):
        super(ManifestValidator, self).__init__(get_schema_path("fov_manifest", doc, version))


LatestCodebookValidator = lambda: CodebookValidator(version=CODEBOOK_CURRENT_VERSION)


LatestExperimentValidator = lambda: ExperimentValidator(version=EXPERIMENT_CURRENT_VERSION)


LatestFOVValidator = lambda: FOVValidator(version=TILESET_CURRENT_VERSION)


LatestManifestValidator = lambda: ManifestValidator(version=TILESET_CURRENT_VERSION)



from ._base import FilterAlgorithm
from .bandpass import Bandpass
from .clip import Clip
from .clip_percentile_to_zero import ClipPercentileToZero
from .clip_value_to_zero import ClipValueToZero
from .element_wise_mult import ElementWiseMultiply
from .gaussian_high_pass import GaussianHighPass
from .gaussian_low_pass import GaussianLowPass
from .laplace import Laplace
from .linear_unmixing import LinearUnmixing
from .map import Map
from .match_histograms import MatchHistograms
from .mean_high_pass import MeanHighPass
from .reduce import Reduce
from .richardson_lucy_deconvolution import DeconvolvePSF
from .white_tophat import WhiteTophat
from .zero_by_channel_magnitude import ZeroByChannelMagnitude

# autodoc's automodule directive only captures the modules explicitly listed in __all__.
all_filters = {
    filter_name: filter_cls
    for filter_name, filter_cls in locals().items()
    if isinstance(filter_cls, type) and issubclass(filter_cls, FilterAlgorithm)
}
__all__ = list(all_filters.keys())



from itertools import product
from typing import Callable

import numpy as np
import pandas as pd

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import (
    Axes,
    Features,
    Number,
    PerImageSliceSpotResults,
    SpotAttributes,
    SpotFindingResults
)


def measure_intensities_at_spot_locations_in_image(
        image: np.ndarray,
        spots: SpotAttributes,
        measurement_function: Callable[[np.ndarray], Number],
        radius_is_gyration: bool = False,
) -> pd.Series:
    """measure the intensity of each spot in spots in the corresponding image

    Parameters
    ----------
    image : np.ndarray,
        3-d volume in which to measure intensities
    spots : pd.DataFrame
        SpotAttributes table containing coordinates and radii of spots
    measurement_function : Callable[[np.ndarray], Number])
        Function to apply over the spot volumes to identify the intensity (e.g. max, mean, ...)
    radius_is_gyration : bool
        if True, indicates that the radius corresponds to radius of gyration, which is a
        function of spot intensity, but typically is a smaller unit than the sigma generated by
        blob_log. In this case, the spot's bounding box is rounded up instead of down when
        measuring intensity. (default False)

    Returns
    -------
    pd.Series :
        Intensities for each spot in SpotAttributes

    """

    def fn(row: pd.Series) -> Number:
        d = image[row['z_min']:row['z_max'], row['y_min']:row['y_max'], row['x_min']:row['x_max']]
        return measurement_function(d)

    if radius_is_gyration:
        radius = np.ceil(spots.data[Features.SPOT_RADIUS]).astype(int) + 1  # round up
    else:
        radius = spots.data[Features.SPOT_RADIUS].astype(int)  # truncate down to nearest int
    for v, max_size in zip(['z', 'y', 'x'], image.shape):
        # numpy does exclusive max indexing, so need to subtract 1 from min to get centered box
        spots.data[f'{v}_min'] = np.clip(spots.data[v] - (radius - 1), 0, None)
        spots.data[f'{v}_max'] = np.clip(spots.data[v] + radius, None, max_size)
    return spots.data[['z_min', 'z_max', 'y_min', 'y_max', 'x_min', 'x_max']].astype(int).apply(
        fn,
        axis=1
    )


def measure_intensities_at_spot_locations_across_imagestack(
        data_image: ImageStack,
        reference_spots: PerImageSliceSpotResults,
        measurement_function: Callable[[np.ndarray], Number],
        radius_is_gyration: bool = False) -> SpotFindingResults:
    """given spots found from a reference image, find those spots across a data_image

    Parameters
    ----------
    data_image : ImageStack
        ImageStack containing multiple volumes for which spots' intensities must be calculated
    reference_spots : PerImageSliceSpotResults
        Spots found in a reference image
    measurement_function : Callable[[np.ndarray], Number])
        Function to apply over the spot volumes to identify the intensity (e.g. max, mean, ...)
    radius_is_gyration : bool
        if True, indicates that the radius corresponds to radius of gyration, which is
        a function of spot intensity, but typically is a smaller unit than the sigma generated
        by blob_log. In this case, the spot's bounding box is rounded up instead of down when
        measuring intensity. (default False)

    Returns
    -------
    SpotFindingResults :
        A Dict of tile indices and their corresponding measured SpotAttributes

    """

    ch_labels = data_image.axis_labels(Axes.CH)
    round_labels = data_image.axis_labels(Axes.ROUND)

    spot_results = SpotFindingResults(imagestack_coords=data_image.xarray.coords,
                                      log=data_image.log)
    # measure spots in each tile
    indices = product(ch_labels, round_labels)
    for c, r in indices:
        tile_indices = {Axes.ROUND: r, Axes.CH: c}
        if reference_spots.spot_attrs.data.empty:
            # if no spots found don't measure
            spot_results[tile_indices] = reference_spots
        else:
            image, _ = data_image.get_slice({Axes.CH: c, Axes.ROUND: r})
            blob_intensities: pd.Series = measure_intensities_at_spot_locations_in_image(
                image,
                reference_spots.spot_attrs,
                measurement_function,
                radius_is_gyration=radius_is_gyration
            )
            # copy reference spot positions and attributes
            tile_spots = SpotAttributes(reference_spots.spot_attrs.data.copy())
            # fill in intensities
            tile_spots.data[Features.INTENSITY] = blob_intensities
            spot_results[tile_indices] = PerImageSliceSpotResults(
                spot_attrs=tile_spots, extras=None)
    return spot_results



import numpy as np

from starfish.core.types import Axes
from .factories import synthetic_stack


def test_get_slice_simple_index():
    """
    Retrieve a slice across one of the axes at the end.  For instance, if the axes are
    (P, Q0,..., Qn-1, R), slice across either P or R.  This test has expectations regarding the
    ordering of the axes in the ImageStack.
    """
    stack = synthetic_stack()
    round_label = 1
    imageslice, axes = stack.get_slice(
        {Axes.ROUND: round_label}
    )
    assert axes == [Axes.CH, Axes.ZPLANE]

    y, x = stack.tile_shape

    for ch_label in range(stack.shape[Axes.CH]):
        for zplane_label in range(stack.shape[Axes.ZPLANE]):
            data = np.empty((y, x))
            data.fill(
                (round_label * stack.shape[Axes.CH] + ch_label) * stack.shape[Axes.ZPLANE]
                + zplane_label)

            assert data.all() == imageslice[ch_label, zplane_label].all()


def test_get_slice_middle_index():
    """
    Retrieve a slice across one of the axes in the middle.  For instance, if the axes are
    (P, Q0,..., Qn-1, R), slice across one of the Q axes.  This test has expectations regarding the
    ordering of the axes in the ImageStack.
    """
    stack = synthetic_stack()
    ch = 1
    imageslice, axes = stack.get_slice(
        {Axes.CH: ch}
    )
    assert axes == [Axes.ROUND, Axes.ZPLANE]

    y, x = stack.tile_shape

    for round_ in range(stack.shape[Axes.ROUND]):
        for z in range(stack.shape[Axes.ZPLANE]):
            data = np.empty((y, x))
            data.fill((round_ * stack.shape[Axes.CH] + ch) * stack.shape[Axes.ZPLANE] + z)

            assert data.all() == imageslice[round_, z].all()


def test_get_slice_range():
    """
    Retrieve a slice across a range of one of the dimensions.
    """
    stack = synthetic_stack()
    zrange = slice(1, 3)
    imageslice, axes = stack.get_slice(
        {Axes.ZPLANE: zrange}
    )
    y, x = stack.tile_shape
    assert axes == [Axes.ROUND, Axes.CH, Axes.ZPLANE]

    for round_ in range(stack.shape[Axes.ROUND]):
        for ch in range(stack.shape[Axes.CH]):
            for z in range(zrange.stop - zrange.start):
                data = np.empty((y, x))
                data.fill((round_ * stack.shape[Axes.CH] + ch) * stack.shape[Axes.ZPLANE]
                          + (z + zrange.start))

                assert data.all() == imageslice[round_, ch, z].all()


def test_set_slice_simple_index():
    """
    Sets a slice across one of the axes at the end.  For instance, if the axes are
    (P, Q0,..., Qn-1, R), sets a slice across either P or R.  This test has expectations regarding
    the ordering of the axes in the ImageStack.
    """
    stack = synthetic_stack()
    round_ = 1
    y, x = stack.tile_shape

    expected = np.full(
        (stack.shape[Axes.CH], stack.shape[Axes.ZPLANE], y, x),
        fill_value=0.5,
        dtype=np.float32
    )
    index = {Axes.ROUND: round_}

    stack.set_slice(index, expected, [Axes.CH, Axes.ZPLANE])

    assert np.array_equal(stack.get_slice(index)[0], expected)


def test_set_slice_middle_index():
    """
    Sets a slice across one of the axes in the middle.  For instance, if the axes are
    (P, Q0,..., Qn-1, R), slice across one of the Q axes.  This test has expectations regarding the
    ordering of the axes in the ImageStack.
    """
    stack = synthetic_stack()
    ch = 1
    y, x = stack.tile_shape

    expected = np.full(
        (stack.shape[Axes.ROUND], stack.shape[Axes.ZPLANE], y, x),
        fill_value=0.5,
        dtype=np.float32
    )
    index = {Axes.CH: ch}

    stack.set_slice(index, expected, [Axes.ROUND, Axes.ZPLANE])

    assert np.array_equal(stack.get_slice(index)[0], expected)


def test_set_slice_reorder():
    """
    Sets a slice across one of the axes.  The source data is not in the same order as the axes in
    ImageStack, but set_slice should reorder the axes and write it correctly.
    """
    stack = synthetic_stack()
    round_ = 1
    y, x = stack.tile_shape
    index = {Axes.ROUND: round_}

    written = np.full(
        (stack.shape[Axes.ZPLANE], stack.shape[Axes.CH], y, x),
        fill_value=0.5,
        dtype=np.float32
    )
    stack.set_slice(index, written, [Axes.ZPLANE, Axes.CH])

    expected = np.full(
        (stack.shape[Axes.CH], stack.shape[Axes.ZPLANE], y, x),
        fill_value=0.5,
        dtype=np.float32
    )
    assert np.array_equal(stack.get_slice(index)[0], expected)


def test_set_slice_range():
    """
    Sets a slice across a range of one of the axes.
    """
    stack = synthetic_stack()
    zrange = slice(1, 3)
    y, x = stack.tile_shape

    expected = np.full(
        (stack.shape[Axes.ROUND], stack.shape[Axes.CH], zrange.stop - zrange.start + 1, y, x),
        fill_value=0.5,
        dtype=np.float32
    )
    index = {Axes.ZPLANE: zrange}

    stack.set_slice(index, expected, [Axes.ROUND, Axes.CH, Axes.ZPLANE])

    assert np.array_equal(stack.get_slice(index)[0], expected)



import os
import time
from pathlib import Path
from typing import Mapping, Sequence

import numpy as np
import pytest
from slicedimage import ImageFormat

from starfish.core.experiment.experiment import Experiment, FieldOfView
from starfish.core.imagestack.test.imagestack_test_utils import verify_physical_coordinates
from starfish.core.types import Axes, Coordinates, CoordinateValue
from .util import (
    render_coordinates_to_rows,
    UniqueTiles,
    write_coordinates_csv,
    write_tile_data,
)
from ..factories import unique_data
from ... import (
    tile_fetcher_factory,
    TileFetcher,
    TileIdentifier,
)
from ...structured_formatter import format_structured_dataset


@pytest.mark.parametrize(
    ["tile_format", "in_place"],
    [
        (ImageFormat.NUMPY, True),
        (ImageFormat.NUMPY, False),
        (ImageFormat.TIFF, True),
        (ImageFormat.TIFF, False),
    ]
)
def test_single_aligned_regular_fov(
        tmpdir,
        tile_format: ImageFormat,
        in_place: bool,
        rounds=(1, 2, 4),
        chs=(2, 3, 4),
        zplanes=(0, 1, 2),
        tile_height=100,
        tile_width=60,
        x_coords=(0.0, 0.1),
        y_coords=(0.1, 0.2),
        zplane_to_coords={0: 0.20, 1: 0.25, 2: 0.3},
        # default value is mutable, but it's for readability reasons!
):
    """Write the tiles for a single aligned (physical coordinates) regular (the dimensions have the
    same cardinality) FOV.  Then build an experiment from the tiles.  Finally, load the resulting
    experiment as an ImageStack and verify that the data matches."""
    tmpdir_path: Path = Path(tmpdir)
    tile_identifiers: Sequence[TileIdentifier] = [
        TileIdentifier(0, round_label, ch_label, zplane_label)
        for round_label in rounds
        for ch_label in chs
        for zplane_label in zplanes
    ]
    tile_fetcher: TileFetcher = tile_fetcher_factory(
        UniqueTiles,
        pass_tile_indices=True,
        fovs=[0],
        rounds=rounds,
        chs=chs,
        zplanes=zplanes,
        tile_height=tile_height,
        tile_width=tile_width,
    )
    tile_coordinates: Mapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]] = {
        tile_identifier: {
            Coordinates.X: x_coords,
            Coordinates.Y: y_coords,
            Coordinates.Z: zplane_to_coords[tile_identifier.zplane_label],
        }
        for tile_identifier in tile_identifiers
    }

    write_tile_data(
        tmpdir_path,
        FieldOfView.PRIMARY_IMAGES,
        tile_format,
        tile_identifiers,
        tile_fetcher)

    coordinates_csv_path = tmpdir_path / "coordinates.csv"
    rows = render_coordinates_to_rows(tile_coordinates)
    write_coordinates_csv(coordinates_csv_path, rows, True)

    # Sleeping by 1 second will result in different timestamps written to TIFF files (the timestamp
    # in the TIFF header has 1 second resolution).  This exposes potential bugs, depending on the
    # nature of the bug and whether in_place is True.
    if tile_format == ImageFormat.TIFF:
        time.sleep(1)

    format_structured_dataset(
        os.fspath(tmpdir_path),
        os.fspath(coordinates_csv_path),
        os.fspath(tmpdir_path),
        tile_format,
        in_place,
    )

    # load the data and verify it.
    exp = Experiment.from_json(os.fspath(tmpdir_path / "experiment.json"))
    fov = exp.fov()
    stack = fov.get_image(FieldOfView.PRIMARY_IMAGES)

    for round_label in rounds:
        for ch_label in chs:
            for zplane_label in zplanes:
                data, _ = stack.get_slice({
                    Axes.ROUND: round_label, Axes.CH: ch_label, Axes.ZPLANE: zplane_label
                })
                expected_data = unique_data(
                    0, rounds.index(round_label), chs.index(ch_label), zplanes.index(zplane_label),
                    1, len(rounds), len(chs), len(zplanes),
                    tile_height, tile_width,
                )
                assert np.allclose(data, expected_data)

    for selectors in stack._iter_axes({Axes.ZPLANE}):
        zplane_label = selectors[Axes.ZPLANE]
        verify_physical_coordinates(
            stack, x_coords, y_coords, zplane_to_coords[zplane_label], selectors[Axes.ZPLANE])


def test_single_ragged_fov(
        tmpdir,
        raw_tile_identifiers=(
            (0, 0, 0),
            (0, 0, 1),
            (0, 0, 2),
            (0, 1, 0),
            (0, 1, 1),
            (0, 1, 2),
            (1, 0, 0),
        ),
        tile_height=100,
        tile_width=60,
        x_coords=(0.0, 0.1),
        y_coords=(0.1, 0.2),
        zplane_to_coords={0: 0.20, 1: 0.25, 2: 0.3},
        # default value is mutable, but it's for readability reasons!
):
    """Write the tiles for a single aligned (physical coordinates) ragged (the dimensions do not
    always have the same cardinality) FOV.  Then build an experiment from the tiles.  Finally, load
    the resulting experiment as an ImageStack and verify that the data matches."""
    tmpdir_path: Path = Path(tmpdir)
    tile_identifiers: Sequence[TileIdentifier] = [
        TileIdentifier(0, round_label, ch_label, zplane_label)
        for round_label, ch_label, zplane_label in raw_tile_identifiers
    ]
    rounds = sorted(set([round_label for round_label, _, _ in raw_tile_identifiers]))
    chs = sorted(set([ch_label for _, ch_label, _ in raw_tile_identifiers]))
    zplanes = sorted(set([zplane_label for _, _, zplane_label in raw_tile_identifiers]))
    tile_fetcher: TileFetcher = tile_fetcher_factory(
        UniqueTiles,
        pass_tile_indices=True,
        fovs=[0],
        rounds=rounds,
        chs=chs,
        zplanes=zplanes,
        tile_height=tile_height,
        tile_width=tile_width,
    )
    tile_coordinates: Mapping[TileIdentifier, Mapping[Coordinates, CoordinateValue]] = {
        tile_identifier: {
            Coordinates.X: x_coords,
            Coordinates.Y: y_coords,
            Coordinates.Z: zplane_to_coords[tile_identifier.zplane_label],
        }
        for tile_identifier in tile_identifiers
    }

    write_tile_data(
        tmpdir_path,
        FieldOfView.PRIMARY_IMAGES,
        ImageFormat.TIFF,
        tile_identifiers,
        tile_fetcher)

    coordinates_csv_path = tmpdir_path / "coordinates.csv"
    rows = render_coordinates_to_rows(tile_coordinates)
    write_coordinates_csv(coordinates_csv_path, rows, True)

    format_structured_dataset(
        os.fspath(tmpdir_path),
        os.fspath(coordinates_csv_path),
        os.fspath(tmpdir_path),
        ImageFormat.TIFF,
        False,
    )

    # load the data and verify it.
    exp = Experiment.from_json(os.fspath(tmpdir_path / "experiment.json"))
    fov = exp.fov()

    for loaded_rounds, expected_chs, expected_zplanes in (
            ({0}, {0, 1}, {0, 1, 2}),
            ({1}, {0}, {0}),
    ):
        stacks = list(fov.get_images(FieldOfView.PRIMARY_IMAGES, rounds=loaded_rounds))

        assert len(stacks) == 1
        stack = stacks[0]
        assert set(stack.axis_labels(Axes.CH)) == set(expected_chs)
        assert set(stack.axis_labels(Axes.ZPLANE)) == set(expected_zplanes)

        for round_label in loaded_rounds:
            for ch_label in expected_chs:
                for zplane_label in expected_zplanes:
                    data, _ = stack.get_slice({
                        Axes.ROUND: round_label, Axes.CH: ch_label, Axes.ZPLANE: zplane_label
                    })
                    expected_data = unique_data(
                        0,
                        rounds.index(round_label), chs.index(ch_label), zplanes.index(zplane_label),
                        1, len(rounds), len(chs), len(zplanes),
                        tile_height, tile_width,
                    )
                    assert np.allclose(data, expected_data)

        for selectors in stack._iter_axes({Axes.ZPLANE}):
            zplane_label = selectors[Axes.ZPLANE]
            verify_physical_coordinates(
                stack, x_coords, y_coords, zplane_to_coords[zplane_label], selectors[Axes.ZPLANE])



import warnings
from collections import OrderedDict

import numpy as np
import pytest

from starfish import data
from starfish import ImageStack
from starfish.core.image.Filter.reduce import Reduce
from starfish.core.imagestack.test.factories import imagestack_with_coords_factory
from starfish.core.imagestack.test.imagestack_test_utils import verify_physical_coordinates
from starfish.types import Axes, FunctionSource, PhysicalCoordinateTypes


X_COORDS = 1, 2
Y_COORDS = 4, 6
Z_COORDS = 1, 3


def make_image_stack():
    """Make a test ImageStack."""

    # Make the test image
    test = np.ones((2, 4, 1, 2, 2), dtype='float32') * 0.1

    x = [0, 0, 1, 1]
    y = [0, 1, 0, 1]

    for i in range(4):
        test[0, i, 0, x[i], y[i]] = 1
    test[0, 0, 0, 0, 0] = 0.75

    # Make the ImageStack
    test_stack = ImageStack.from_numpy(test)

    return test_stack


def make_expected_image_stack(func):
    """Make the expected image stack result"""

    if func == 'max':
        reduced = np.array(
            [[[[[0.75, 0.1],
                [0.1, 0.1]]],
                [[[0.1, 1],
                  [0.1, 0.1]]],
                [[[0.1, 0.1],
                  [1, 0.1]]],
                [[[0.1, 0.1],
                  [0.1, 1]]]]], dtype='float32'
        )
    elif func == 'mean':
        reduced = np.array(
            [[[[[0.425, 0.1],
              [0.1, 0.1]]],
              [[[0.1, 0.55],
                [0.1, 0.1]]],
              [[[0.1, 0.1],
                [0.55, 0.1]]],
              [[[0.1, 0.1],
                [0.1, 0.55]]]]], dtype='float32'
        )
    elif func == 'sum':
        reduced = np.array(
            [[[[[0.85, 0.2],
                [0.2, 0.2]]],
              [[[0.2, 1],
                [0.2, 0.2]]],
              [[[0.2, 0.2],
                [1, 0.2]]],
              [[[0.2, 0.2],
                [0.2, 1]]]]], dtype='float32'
        )
    elif func == 'norm':
        reduced = np.array(
            [[[[[0.75663730, 0.14142136],
                [0.14142136, 0.14142136]]],
              [[[0.14142136, 1.00000000],
                [0.14142136, 0.14142136]]],
              [[[0.14142136, 0.14142136],
                [1.00000000, 0.14142136]]],
              [[[0.14142136, 0.14142136],
                [0.14142136, 1.00000000]]]]],
            dtype=np.float32
        )
    else:
        raise ValueError("Unsupported func")

    expected_stack = ImageStack.from_numpy(reduced)

    return expected_stack


@pytest.mark.parametrize(
    "expected_result,func,module,kwargs",
    [
        (make_expected_image_stack('max'), 'max', None, None),
        (make_expected_image_stack('mean'), 'mean', None, None),
        (make_expected_image_stack('sum'), 'sum', None, None),
        (
            make_expected_image_stack('norm'),
            'linalg.norm',
            FunctionSource.scipy,
            {'ord': 2},
        ),
        (
            make_expected_image_stack('norm'),
            FunctionSource.scipy('linalg.norm'),
            None,
            {'ord': 2},
        ),
    ]
)
def test_image_stack_reduce(expected_result, func, module, kwargs):

    # Get the test stack and expected result
    test_stack = make_image_stack()

    # Filter
    if kwargs is None:
        actual_kwargs = {}
    else:
        actual_kwargs = kwargs.copy()
    if module is not None:
        actual_kwargs['module'] = module
    red = Reduce(dims=[Axes.ROUND], func=func, **actual_kwargs)
    reduced = red.run(test_stack)

    assert np.allclose(reduced.xarray, expected_result.xarray)


def test_image_stack_module_deprecated():
    """Specifying the function as a string and passing in a module should generate a warning."""
    with warnings.catch_warnings(record=True) as all_warnings:
        Reduce(dims=[Axes.ROUND], func="max", module=FunctionSource.np)

        assert DeprecationWarning in (warning.category for warning in all_warnings)


def test_image_stack_module_with_functionsourcebundle():
    """Specifying the function as a FunctionSourceBundle and passing in a module should raise an
    Exception."""
    with pytest.raises(ValueError):
        Reduce(dims=[Axes.ROUND], func=FunctionSource.np("max"), module=FunctionSource.np)


def test_max_projection_preserves_coordinates():
    e = data.ISS(use_test_data=True)
    nuclei = e.fov().get_image('nuclei')

    red = Reduce(dims=[Axes.ROUND, Axes.CH, Axes.ZPLANE], func='max')
    nuclei_proj = red.run(nuclei)

    # Since this data already has only 1 round, 1 ch, 1 zplane
    # let's just assert that the max_proj operation didn't change anything
    assert nuclei.xarray.equals(nuclei_proj.xarray)

    stack_shape = OrderedDict([(Axes.ROUND, 3), (Axes.CH, 2),
                               (Axes.ZPLANE, 3), (Axes.Y, 10), (Axes.X, 10)])

    # Create stack with coordinates, verify coords unaffected by max_poj
    physical_coords = OrderedDict([(PhysicalCoordinateTypes.X_MIN, X_COORDS[0]),
                                   (PhysicalCoordinateTypes.X_MAX, X_COORDS[1]),
                                   (PhysicalCoordinateTypes.Y_MIN, Y_COORDS[0]),
                                   (PhysicalCoordinateTypes.Y_MAX, Y_COORDS[1]),
                                   (PhysicalCoordinateTypes.Z_MIN, Z_COORDS[0]),
                                   (PhysicalCoordinateTypes.Z_MAX, Z_COORDS[1])])

    stack = imagestack_with_coords_factory(stack_shape, physical_coords)

    stack_proj = red.run(stack)
    expected_z = np.average(Z_COORDS)
    verify_physical_coordinates(stack_proj, X_COORDS, Y_COORDS, expected_z)



import warnings
from typing import Callable, Collection, Mapping, Sequence

import pytest

from starfish.core.errors import DataFormatWarning
from starfish.core.types import Coordinates, CoordinateValue
from .factories.all_purpose import collection_factory
from .factories.unique_tiles import UniqueTiles
from .. import TileIdentifier


@pytest.mark.parametrize(
    "num_round_labels, num_ch_labels, num_zplane_labels, tile_height, tile_width, has_warnings",
    (
        (1, 1, 1, 100, 100, False),
        (1, 1, 1000, 100, 100, True),
        (1, 1000, 1, 100, 100, True),
        (1000, 1, 1, 100, 100, True),
        (1000, 1, 1, 10000, 100, True),
        (1000, 1, 1, 100, 10000, True),
    )
)
def test_small_axis(
        num_round_labels,
        num_ch_labels,
        num_zplane_labels,
        tile_height,
        tile_width,
        has_warnings,
):
    tile_identifiers: Collection[TileIdentifier] = [
        TileIdentifier(0, round_label, ch_label, zplane_label)
        for round_label in range(num_round_labels)
        for ch_label in range(num_ch_labels)
        for zplane_label in range(num_zplane_labels)
    ]

    def make_tile_coordinate_callback(
            all_zplane_labels: Sequence[int]
    ) -> Callable[[TileIdentifier], Mapping[Coordinates, CoordinateValue]]:
        def tile_coordinate_callback(
                tile_identifier: TileIdentifier
        ) -> Mapping[Coordinates, CoordinateValue]:
            zplane_offset = all_zplane_labels.index(tile_identifier.zplane_label)
            return {
                Coordinates.X: (0.0, 0.1),
                Coordinates.Y: (0.0, 0.1),
                Coordinates.Z: zplane_offset * 0.1,
            }

        return tile_coordinate_callback

    with warnings.catch_warnings(record=True) as warnings_:
        collection_factory(
            UniqueTiles,
            tile_identifiers,
            make_tile_coordinate_callback(
                sorted(set(tile_identifier.zplane_label for tile_identifier in tile_identifiers))),
            tile_height,
            tile_width,
        )
        assert has_warnings == (
            0 != len(
                [warning for warning in warnings_
                 if issubclass(warning.category, DataFormatWarning)]))



from starfish.core.imagestack.test.factories import synthetic_stack
from starfish.core.types import FunctionSource
from .. import Map


def test_map():
    """test that apply correctly applies a simple function across 2d tiles of a Stack"""
    stack = synthetic_stack()
    assert (stack.xarray == 1).all()
    mapper = Map("divide", 2)
    output = mapper.run(stack)
    assert (output.xarray == 0.5).all()

    mapper = Map(FunctionSource.np("divide"), 2)
    output = mapper.run(stack)
    assert (output.xarray == 0.5).all()



from abc import ABCMeta
from typing import Callable, cast, Collection, Mapping, Sequence, Type, Union

import numpy as np
import slicedimage

from starfish.core.experiment.builder import (
    build_irregular_image,
    FetchedTile,
    tile_fetcher_factory,
    TileFetcher,
    TileIdentifier,
)
from starfish.core.types import Axes, Coordinates, CoordinateValue


class LocationAwareFetchedTile(FetchedTile, metaclass=ABCMeta):
    """This is the base class for tiles that are aware of their location in the 5D tensor.
    """
    def __init__(
            self,
            # these are the arguments passed in as a result of tile_fetcher_factory's
            # pass_tile_indices parameter.
            fov_id: int, round_label: int, ch_label: int, zplane_label: int,
            # these are the arguments we are passing through tile_fetcher_factory.
            fovs: Sequence[int], rounds: Sequence[int], chs: Sequence[int], zplanes: Sequence[int],
            tile_height: int, tile_width: int,
    ) -> None:
        super().__init__()
        self.fov_id = fov_id
        self.round_label = round_label
        self.ch_label = ch_label
        self.zplane_label = zplane_label
        self.fovs = fovs
        self.rounds = rounds
        self.chs = chs
        self.zplanes = zplanes
        self.tile_height = tile_height
        self.tile_width = tile_width


def _apply_coords_range_fetcher(
        backing_tile_fetcher: TileFetcher,
        tile_coordinates_callback: Callable[
            [TileIdentifier], Mapping[Coordinates, CoordinateValue]],
) -> TileFetcher:
    """Given a :py:class:`TileFetcher`, intercept all the returned :py:class:`FetchedTile` instances
    and replace the coordinates using the coordinates from `tile_coordinates_callback`."""
    class ModifiedTile(FetchedTile):
        def __init__(
                self,
                backing_tile: FetchedTile,
                tile_identifier: TileIdentifier,
                *args, **kwargs
        ):
            super().__init__(*args, **kwargs)
            self.backing_tile = backing_tile
            self.tile_identifier = tile_identifier

        @property
        def shape(self) -> Mapping[Axes, int]:
            return self.backing_tile.shape

        @property
        def coordinates(self) -> Mapping[Union[str, Coordinates], CoordinateValue]:
            return cast(
                Mapping[Union[str, Coordinates], CoordinateValue],
                tile_coordinates_callback(self.tile_identifier))

        def tile_data(self) -> np.ndarray:
            return self.backing_tile.tile_data()

    class ModifiedTileFetcher(TileFetcher):
        def get_tile(
                self, fov_id: int, round_label: int, ch_label: int, zplane_label: int,
        ) -> FetchedTile:
            original_fetched_tile = backing_tile_fetcher.get_tile(
                fov_id, round_label, ch_label, zplane_label)
            tile_identifier = TileIdentifier(fov_id, round_label, ch_label, zplane_label)
            return ModifiedTile(original_fetched_tile, tile_identifier)

    return ModifiedTileFetcher()


def collection_factory(
        fetched_tile_cls: Type[LocationAwareFetchedTile],
        tile_identifiers: Collection[TileIdentifier],
        tile_coordinates_callback: Callable[
            [TileIdentifier], Mapping[Coordinates, CoordinateValue]],
        tile_height: int,
        tile_width: int,
) -> slicedimage.Collection:
    """Given a type that implements the :py:class:`LocationAwareFetchedTile` contract, produce a
    slicedimage Collection with the tiles in `tile_identifiers`.  For a given tile_identifier,
    retrieve the coordinates by invoking the callback `tile_coordinates_callback`.

    Parameters
    ----------
    fetched_tile_cls : Type[LocationAwareFetchedTile]
        The class of the FetchedTile.
    tile_identifiers : Collection[TileIdentifier]
        TileIdentifiers for each of the tiles in the collection.
    tile_coordinates_callback : Callable[[TileIdentifier], Mapping[Coordinates, CoordinatesValue]]
        A callable that returns the coordinates for a given tile's TileIdentifier.
    tile_height : int
        Height of each tile, in pixels.
    tile_width : int
        Width of each tile, in pixels.
    """
    all_fov_ids = sorted(set(
        tile_identifier.fov_id for tile_identifier in tile_identifiers))
    all_round_labels = sorted(set(
        tile_identifier.round_label for tile_identifier in tile_identifiers))
    all_ch_labels = sorted(set(
        tile_identifier.ch_label for tile_identifier in tile_identifiers))
    all_zplane_labels = sorted(set(
        tile_identifier.zplane_label for tile_identifier in tile_identifiers))

    original_tile_fetcher = tile_fetcher_factory(
        fetched_tile_cls, True,
        all_fov_ids, all_round_labels, all_ch_labels, all_zplane_labels,
        tile_height, tile_width,
    )
    modified_tile_fetcher = _apply_coords_range_fetcher(
        original_tile_fetcher, tile_coordinates_callback)

    return build_irregular_image(
        tile_identifiers,
        modified_tile_fetcher,
        default_shape={Axes.Y: tile_height, Axes.X: tile_width}
    )



import json
import platform
from functools import lru_cache
from typing import List, Mapping

import pkg_resources

# these are import statements and not from xxx import yyy to break a circular dependency.
import starfish.core
import starfish.core.types


class Log:
    def __init__(self):
        """
        Class for capturing methods and their parameters applied by an analysis
        pipeline.
        """
        self._log: List[dict] = list()

    def update_log(self, class_instance) -> None:
        """
        Adds a new entry to the log list.

        Parameters
        ----------
        class_instance: The instance of a class being applied to the imagestack
        """
        entry = {"method": class_instance.__class__.__name__,
                 "arguments": class_instance.__dict__,
                 "os": get_os_info(),
                 "dependencies": get_core_dependency_info(),
                 "release tag": get_release_tag(),
                 "starfish version": get_dependency_version('starfish')
                 }
        self._log.append(entry)

    def encode(self):
        return LogEncoder().encode({starfish.core.types.LOG: self._log})

    @classmethod
    def decode(cls, encoded_log: str):
        log = json.loads(encoded_log)
        log_object = Log()
        log_object._log = log
        return log_object

    @property
    def data(self):
        return self._log


@lru_cache(maxsize=1)
def get_core_dependency_info() -> Mapping[str, str]:
    dependency_info = dict()
    for dependency in starfish.core.types.CORE_DEPENDENCIES:
        version = get_dependency_version(dependency)
        dependency_info[dependency] = version
    return dependency_info


def get_dependency_version(dependency: str) -> str:
    return pkg_resources.get_distribution(dependency).version


@lru_cache(maxsize=1)
def get_release_tag() -> str:
    if not starfish.core.is_release_tag:
        return "Running starfish from source"
    return starfish.core.is_release_tag


@lru_cache(maxsize=1)
def get_os_info() -> Mapping[str, str]:
    return {"Platform": platform.system(),
            "Version:": platform.version(),
            "Python Version": platform.python_version()}


class LogEncoder(json.JSONEncoder):
    """
    JSON encodes the List[Dict] pipeline provence log. For simple
    objects use default JSON encoding. For more complex objects
    (ex. Imagestack, Codebook) encode the repr of the object.
    """
    def default(self, o):
        try:
            return super(LogEncoder, self).default(o)
        except TypeError:
            return json.JSONEncoder().encode(repr(o))



from typing import Sequence, Tuple

import numpy as np
from scipy.ndimage import gaussian_filter

from starfish.core.codebook.codebook import Codebook
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.types import Axes, Features


def create_imagestack_from_codebook(
    pixel_dimensions: Tuple[int, int, int],
    spot_coordinates: Sequence[Tuple[int, int, int]],
    codebook: Codebook
) -> ImageStack:
    """
    creates a numpy array containing one spot per codebook entry at spot_coordinates. length of
    spot_coordinates must therefore match the number of codes in Codebook.
    """
    assert len(spot_coordinates) == codebook.sizes[Features.TARGET]

    data_shape = (
        codebook.sizes[Axes.ROUND.value],
        codebook.sizes[Axes.CH.value],
        *pixel_dimensions
    )
    imagestack_data = np.zeros(data_shape, dtype=np.float32)

    for ((z, y, x), f) in zip(spot_coordinates, range(codebook.sizes[Features.TARGET])):
        imagestack_data[:, :, z, y, x] = codebook[f].transpose(Axes.ROUND.value, Axes.CH.value)

    # blur with a small non-isotropic kernel TODO make kernel smaller.
    imagestack_data = gaussian_filter(imagestack_data, sigma=(0, 0, 0.7, 1.5, 1.5))
    return ImageStack.from_numpy(imagestack_data)



from ._base import AssignTargetsAlgorithm
from .label import Label

# autodoc's automodule directive only captures the modules explicitly listed in __all__.
all_filters = {
    filter_name: filter_cls
    for filter_name, filter_cls in locals().items()
    if isinstance(filter_cls, type) and issubclass(filter_cls, AssignTargetsAlgorithm)
}
__all__ = list(all_filters.keys())



from abc import abstractmethod
from typing import Callable, Optional

import numpy as np

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.types import Number, SpotFindingResults


class FindSpotsAlgorithm(metaclass=AlgorithmBase):
    """
    Starfish spot finders use a variety of means to detect bright spots against
    dark backgrounds. Starfish's spot detectors each have different strengths and weaknesses.

    **Fixed-position spot finders**

    The following spot finders have two modes of operation.

    The first mode is suitable for coded
    experiments where genes are identified by patterns of spots over all rounds and channels of the
    experiment. In this mode, the spot finders identify spots in a single reference image,
    which can be either a dots auxiliary image, or a maximum intensity projection of the primary
    images. The positions of the maxima are then measured in all other images, and the intensities
    across the complete experiment are stored in an :ref:`IntensityTable`

    The second mode is suitable for assays that detect spots in a single round, such as single
    molecule FISH and RNAscope. This mode simply finds all the spots and concatenates them into a
    long-form IntensityTable. In this mode, the spots are not measured in images that correspond to
    other :code:`(round, channel)` pairs; those positions of the IntensityTable are filled with
    :code:`np.nan`.

    1. The :py:class:`~starfish.spots._find_spots.blob.BlobDetector` allows the user to pre-filter
    an image using either a Laplacian-of-Gaussians or
    Difference-of-Gaussians (fast approximation to Laplacian-of-Gaussians). These filters are
    applied at with a user-specified variety of Gaussian kernel sizes, and the best-fitting size is
    automatically selected. This allows this filter to detect Gaussian shaped blobs of various
    sizes.

    """

    @abstractmethod
    def run(self, image_stack: ImageStack,
            reference_image: Optional[ImageStack] = None, *args) -> SpotFindingResults:
        """Find and measure spots across rounds and channels in the provided ImageStack."""
        raise NotImplementedError()

    @staticmethod
    def _get_measurement_function(
            measurement_type: str
    ) -> Callable[[np.ndarray], Number]:
        try:
            measurement_function = getattr(np, measurement_type)
        except AttributeError:
            raise ValueError(
                f'measurement_type must be a numpy reduce function such as "max" or "mean". '
                f'{measurement_type} not found.')
        return measurement_function



import numpy as np
import pytest
import xarray as xr

from ..levels import preserve_float_range


def combos(make_xarrays: bool):
    test_parameters = (
        # none of these will change the data
        (
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
        ),

        # different dtype.
        (
            np.asarray((0.0, 0.0, 1.0), dtype=np.float64),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
        ),

        # some below zero stuff
        (
            np.asarray((-1.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
        ),

        # some above one stuff
        (
            np.asarray((0.0, 1.0, 2.0), dtype=np.float32),
            np.asarray((0.0, 1.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.5, 1.0), dtype=np.float32),
        ),

        # both below zero and above one stuff
        (
            np.asarray((-1.0, 0.0, 1.0, 2.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
        ),

        # both below zero and above one stuff, different dtype
        (
            np.asarray((-1.0, 0.0, 1.0, 2.0), dtype=np.float64),
            np.asarray((0.0, 0.0, 1.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
        ),

        # both below zero and above one stuff
        (
            np.asarray((-1, 0, 1, 2), dtype=np.int16),
            np.asarray((0.0, 0.0, 1.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
        ),
    )

    if make_xarrays:
        return (
            (xr.DataArray(source),
             xr.DataArray(expected_clipped_output),
             xr.DataArray(expected_rescaled_output))
            for source, expected_clipped_output, expected_rescaled_output in test_parameters
        )
    else:
        return test_parameters


@pytest.mark.parametrize(
    ("_source", "expected_clipped_output", "expected_rescaled_output"),
    combos(True)
)
def test_preserve_range_xarray(
        _source: xr.DataArray,
        expected_clipped_output: xr.DataArray,
        expected_rescaled_output: xr.DataArray,
):
    def run(source: xr.DataArray, expected_output: xr.DataArray, rescale: bool):
        original_data = source.copy(deep=True)

        # compute while trying to preserve the input.  if there is no change to the data, we should
        # come back with the same block of memory.
        output_preserve_input = preserve_float_range(source, rescale=rescale, preserve_input=True)
        assert expected_output.equals(output_preserve_input)
        assert original_data.equals(source)
        if source.dtype == np.float32 and expected_output.equals(source):
            assert output_preserve_input.values is source.values

        # compute it while trying to save memory.
        output_stingy_memory = preserve_float_range(source, rescale=rescale, preserve_input=False)
        assert expected_output.equals(output_stingy_memory)
        if source.dtype == np.float32:
            assert output_stingy_memory.values is source.values

    run(_source.copy(deep=True), expected_clipped_output, False)
    run(_source.copy(deep=True), expected_rescaled_output, True)


@pytest.mark.parametrize(
    ("_source", "expected_clipped_output", "expected_rescaled_output"),
    combos(False)
)
def test_preserve_range_ndarray(
        _source: np.ndarray,
        expected_clipped_output: np.ndarray,
        expected_rescaled_output: np.ndarray,
):
    def run(source: np.ndarray, expected_output: np.ndarray, rescale: bool):
        original_data = source.copy()

        # compute while trying to preserve the input.  if there is no change to the data, we should
        # come back with the same block of memory.
        output_preserve_input = preserve_float_range(source, rescale=rescale, preserve_input=True)
        assert np.all(np.equal(expected_output, output_preserve_input))
        assert np.all(np.equal(original_data, source))
        if source.dtype == np.float32 and np.all(np.equal(expected_output, source)):
            # no change
            assert output_preserve_input is source

        # compute it while trying to save memory.
        output_stingy_memory = preserve_float_range(source, rescale=rescale, preserve_input=False)
        assert np.all(np.equal(expected_output, output_stingy_memory))
        if source.dtype == np.float32:
            assert output_stingy_memory is source

    run(_source.copy(), expected_clipped_output, False)
    run(_source.copy(), expected_rescaled_output, True)



from starfish.core.morphology import (  # noqa: F401
    Binarize,
    Filter,
)
from starfish.core.morphology.binary_mask import BinaryMaskCollection  # noqa: F401
from starfish.core.morphology.label_image import LabelImage  # noqa: F401



from typing import Callable, Mapping

import pandas as pd

from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.types import (
    Axes,
    Features,
    SpotAttributes,
    SpotFindingResults,
    TraceBuildingStrategies
)
from .util import _build_intensity_table, _match_spots, _merge_spots_by_round


def build_spot_traces_exact_match(spot_results: SpotFindingResults, **kwargs) -> IntensityTable:
    """
    Combines spots found in matching x/y positions across rounds and channels of
    an ImageStack into traces represented as an IntensityTable.

    Parameters
    -----------
    spot_results: SpotFindingResults
        Spots found across rounds/channels of an ImageStack
    """
    # create IntensityTable with same x/y/z info accross all r/ch
    spot_attributes = list(spot_results.values())[0].spot_attrs
    intensity_table = IntensityTable.zeros(
        spot_attributes=spot_attributes,
        round_labels=spot_results.round_labels,
        ch_labels=spot_results.ch_labels,
    )
    for r, c in spot_results.keys():
        value = spot_results[{Axes.ROUND: r, Axes.CH: c}].spot_attrs.data[Features.INTENSITY]
        # if no exact match set value to 0
        value = 0 if value.empty else value
        intensity_table.loc[dict(c=c, r=r)] = value
    return intensity_table


def build_traces_sequential(spot_results: SpotFindingResults, **kwargs) -> IntensityTable:
    """
    Build spot traces  without merging across channels and imaging rounds. Used for sequential
    methods like smFIsh.

    Parameters
    ----------
    spot_results: SpotFindingResults
        Spots found across rounds/channels of an ImageStack

    Returns
    -------
    IntensityTable :
        concatenated input SpotAttributes, converted to an IntensityTable object

    """

    all_spots = pd.concat([sa.spot_attrs.data for sa in spot_results.values()], sort=True)

    intensity_table = IntensityTable.zeros(
        spot_attributes=SpotAttributes(all_spots),
        ch_labels=spot_results.ch_labels,
        round_labels=spot_results.round_labels,
    )

    i = 0
    for (r, c), spot_attrs in spot_results.items():
        for _, row in spot_attrs.spot_attrs.data.iterrows():
            selector = dict(features=i, c=c, r=r)
            intensity_table.loc[selector] = row[Features.INTENSITY]
            i += 1
    return intensity_table


def build_traces_nearest_neighbors(spot_results: SpotFindingResults, anchor_round: int=1,
                                   search_radius: int=3):
    """
    Combine spots found across round and channels of an ImageStack using a nearest neighbors
    strategy

    Parameters
    -----------
    spot_results : SpotFindingResults
        Spots found across rounds/channels of an ImageStack
    anchor_round : int
        The imaging round against which other rounds will be checked for spots in the same
        approximate pixel location.
    search_radius : int
        Number of pixels over which to search for spots in other rounds and channels.

    """
    per_round_spot_results = _merge_spots_by_round(spot_results)

    distances, indices = _match_spots(
        per_round_spot_results,
        anchor_round=anchor_round
    )
    intensity_table = _build_intensity_table(
        per_round_spot_results, distances, indices,
        rounds=spot_results.round_labels,
        channels=spot_results.ch_labels,
        search_radius=search_radius,
        anchor_round=anchor_round
    )
    return intensity_table

TRACE_BUILDERS: Mapping[TraceBuildingStrategies, Callable] = {
    TraceBuildingStrategies.EXACT_MATCH: build_spot_traces_exact_match,
    TraceBuildingStrategies.NEAREST_NEIGHBOR: build_traces_nearest_neighbors,
    TraceBuildingStrategies.SEQUENTIAL: build_traces_sequential,
}



from typing import Mapping, Optional, Type

import numpy as np
import pytest

from starfish import Log
from starfish.image import Filter
from starfish.types import ArrayLike, Axes, Coordinates, Number
from ..label_image import AttrKeys, CURRENT_VERSION, DOCTYPE_STRING, LabelImage


@pytest.mark.parametrize(
    "array, physical_ticks, log, expected_error",
    [
        # 3D label image
        [
            np.zeros((1, 1, 1), dtype=np.int32),
            {
                Coordinates.X: [0],
                Coordinates.Y: [0],
                Coordinates.Z: [0],
            },
            None,
            None,
        ],
        # 2D label image
        [
            np.zeros((1, 2), dtype=np.int32),
            {
                Coordinates.X: [0, 1],
                Coordinates.Y: [0],
            },
            None,
            None,
        ],
        # wrong dtype
        [
            np.zeros((1, 2), dtype=np.float32),
            {
                Coordinates.X: [0, 1],
                Coordinates.Y: [0],
            },
            None,
            TypeError,
        ],
        # missing some coordinates
        [
            np.zeros((1, 2), dtype=np.float32),
            {
                Coordinates.X: [0, 1],
            },
            None,
            KeyError,
        ],
    ]
)
def test_from_array_and_coords(
        array: np.ndarray,
        physical_ticks: Mapping[Coordinates, ArrayLike[Number]],
        log: Optional[Log],
        expected_error: Optional[Type[Exception]],
):
    """Test that we can construct a LabelImage and that some common error conditions are caught."""
    if expected_error is not None:
        with pytest.raises(expected_error):
            LabelImage.from_label_array_and_ticks(array, None, physical_ticks, log)
    else:
        label_image = LabelImage.from_label_array_and_ticks(array, None, physical_ticks, log)
        assert isinstance(label_image.log, Log)
        assert label_image.xarray.attrs.get(AttrKeys.DOCTYPE, None) == DOCTYPE_STRING
        assert label_image.xarray.attrs.get(AttrKeys.VERSION, None) == str(CURRENT_VERSION)


def test_pixel_coordinates():
    """Test that the code creates missing pixel coordinate values."""
    array = np.zeros((2, 3, 4), dtype=np.int32)
    pixel_coordinates = {
        Axes.X: [2, 3, 4, 5],
        Axes.ZPLANE: [0, 1],
    }
    physical_coordinates = {
        Coordinates.X: [0, 0.5, 1.0, 1.5],
        Coordinates.Y: [0, 0.2, 0.4],
        Coordinates.Z: [0, 0.1],
    }
    label_image = LabelImage.from_label_array_and_ticks(
        array, pixel_coordinates, physical_coordinates, None)

    assert np.array_equal(label_image.xarray.coords[Axes.X.value], [2, 3, 4, 5])
    # not provided, should be 0..N-1
    assert np.array_equal(label_image.xarray.coords[Axes.Y.value], [0, 1, 2])
    assert np.array_equal(label_image.xarray.coords[Axes.ZPLANE.value], [0, 1])


def test_coordinates_key_type():
    """Test that the code correctly handles situations where the coordinate keys are provided as
    strings instead of the enumerated types."""
    array = np.zeros((2, 3, 4), dtype=np.int32)
    pixel_coordinates = {
        Axes.X.value: [2, 3, 4, 5],
        Axes.Y.value: [0, 1, 2],
        Axes.ZPLANE.value: [0, 1],
    }
    physical_coordinates = {
        Coordinates.X.value: [0, 0.5, 1.0, 1.5],
        Coordinates.Y.value: [0, 0.2, 0.4],
        Coordinates.Z.value: [0, 0.1],
    }
    label_image = LabelImage.from_label_array_and_ticks(
        array, pixel_coordinates, physical_coordinates, None)

    for axis_str, axis_data in pixel_coordinates.items():
        assert np.array_equal(label_image.xarray.coords[axis_str], axis_data)
    for coord_str, coord_data in physical_coordinates.items():
        assert np.array_equal(label_image.xarray.coords[coord_str], coord_data)


def test_save_and_load(tmp_path):
    """Verify that we can save the label image and load it correctly."""
    array = np.zeros((2, 3, 4), dtype=np.int32)
    pixel_coordinates = {
        Axes.X: [2, 3, 4, 5],
        Axes.ZPLANE: [0, 1],
    }
    physical_coordinates = {
        Coordinates.X: [0, 0.5, 1.0, 1.5],
        Coordinates.Y: [0, 0.2, 0.4],
        Coordinates.Z: [0, 0.1],
    }
    log = Log()
    # instantiate a filter (even though that makes no sense in this context)
    filt = Filter.Reduce((Axes.ROUND,), func="max")
    log.update_log(filt)

    label_image = LabelImage.from_label_array_and_ticks(
        array, pixel_coordinates, physical_coordinates, log)
    label_image.to_netcdf(tmp_path / "label_image.netcdf")

    loaded_label_image = LabelImage.open_netcdf(tmp_path / "label_image.netcdf")

    assert label_image.xarray.equals(loaded_label_image.xarray)
    assert label_image.xarray.attrs == loaded_label_image.xarray.attrs



import numpy as np
import pytest
import xarray as xr

from ..levels import levels


def combos(make_xarrays: bool):
    test_parameters = (
        # none of these will change the data
        (
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
        ),

        # different dtype.
        (
            np.asarray((0.0, 0.0, 1.0), dtype=np.float64),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
        ),

        # some below zero stuff
        (
            np.asarray((-1.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0), dtype=np.float32),
        ),

        # some above one stuff
        (
            np.asarray((0.0, 1.0, 2.0), dtype=np.float32),
            np.asarray((0.0, 1.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.5, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.5, 1.0), dtype=np.float32),
        ),

        # peak below one
        (
            np.asarray((0.0, 0.25, 0.5), dtype=np.float32),
            np.asarray((0.0, 0.25, 0.5), dtype=np.float32),
            np.asarray((0.0, 0.5, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.25, 0.5), dtype=np.float32),
        ),

        # both below zero and above one stuff
        (
            np.asarray((-1.0, 0.0, 1.0, 2.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 1.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
        ),

        # both below zero and above one stuff, different dtype
        (
            np.asarray((-1.0, 0.0, 1.0, 2.0), dtype=np.float64),
            np.asarray((0.0, 0.0, 1.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
        ),

        # both below zero and above one stuff
        (
            np.asarray((-1, 0, 1, 2), dtype=np.int16),
            np.asarray((0.0, 0.0, 1.0, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
            np.asarray((0.0, 0.0, 0.5, 1.0), dtype=np.float32),
        ),
    )

    if make_xarrays:
        return (
            (xr.DataArray(source),
             xr.DataArray(expected_clip_output),
             xr.DataArray(expected_rescale_output),
             xr.DataArray(expected_rescale_saturated_output))
            for (source,
                 expected_clip_output,
                 expected_rescale_output,
                 expected_rescale_saturated_output) in test_parameters
        )
    else:
        return test_parameters


@pytest.mark.parametrize(
    ("_source",
     "expected_clip_output",
     "expected_rescale_output",
     "expected_rescale_saturated_output",
     ),
    combos(True)
)
def test_levels_xarray(
        _source: xr.DataArray,
        expected_clip_output: xr.DataArray,
        expected_rescale_output: xr.DataArray,
        expected_rescale_saturated_output: xr.DataArray,
):
    def run(
            source: xr.DataArray,
            expected_output: xr.DataArray,
            rescale: bool,
            rescale_saturated: bool,
    ):
        original_data = source.copy(deep=True)

        # compute while trying to preserve the input.  if there is no change to the data, we should
        # come back with the same block of memory.
        output_preserve_input = levels(
            source, rescale=rescale, rescale_saturated=rescale_saturated, preserve_input=True)
        assert expected_output.equals(output_preserve_input)
        assert original_data.equals(source)
        if source.dtype == np.float32 and expected_output.equals(source) and not rescale:
            assert output_preserve_input.values is source.values

        # compute it while trying to save memory.
        output_stingy_memory = levels(
            source, rescale=rescale, rescale_saturated=rescale_saturated, preserve_input=False)
        assert expected_output.equals(output_stingy_memory)
        if source.dtype == np.float32:
            assert output_stingy_memory.values is source.values

    run(_source.copy(deep=True), expected_clip_output, False, False)
    run(_source.copy(deep=True), expected_rescale_output, True, False)
    run(_source.copy(deep=True), expected_rescale_saturated_output, False, True)


@pytest.mark.parametrize(
    ("_source",
     "expected_clip_output",
     "expected_rescale_output",
     "expected_rescale_saturated_output",
     ),
    combos(False)
)
def test_levels_ndarray(
        _source: np.ndarray,
        expected_clip_output: np.ndarray,
        expected_rescale_output: np.ndarray,
        expected_rescale_saturated_output: np.ndarray,
):
    def run(
            source: np.ndarray,
            expected_output: np.ndarray,
            rescale: bool,
            rescale_saturated: bool,
    ):
        original_data = source.copy()

        # compute while trying to preserve the input.  if there is no change to the data, we should
        # come back with the same block of memory.
        output_preserve_input = levels(
            source, rescale=rescale, rescale_saturated=rescale_saturated, preserve_input=True)
        assert np.all(np.equal(expected_output, output_preserve_input))
        assert np.all(np.equal(original_data, source))
        if source.dtype == np.float32 and np.all(np.equal(expected_output, source)) and not rescale:
            # no change
            assert output_preserve_input is source

        # compute it while trying to save memory.
        output_stingy_memory = levels(
            source, rescale=rescale, rescale_saturated=rescale_saturated, preserve_input=False)
        assert np.all(np.equal(expected_output, output_stingy_memory))
        if source.dtype == np.float32:
            assert output_stingy_memory is source

    run(_source.copy(), expected_clip_output, False, False)
    run(_source.copy(), expected_rescale_output, True, False)
    run(_source.copy(), expected_rescale_saturated_output, False, True)



from abc import abstractmethod

from starfish.core.image._registration.transforms_list import TransformsList
from starfish.core.pipeline.algorithmbase import AlgorithmBase


class LearnTransformAlgorithm(metaclass=AlgorithmBase):

    @abstractmethod
    def run(self, stack, *args) -> TransformsList:
        """Learns Transforms for a given stack."""
        raise NotImplementedError()



# constants
from starfish.core.types import (  # noqa: F401
    ArrayLike,
    Axes,
    Clip,
    Coordinates,
    CoordinateValue,
    CORE_DEPENDENCIES,
    Features,
    FunctionSource,
    FunctionSourceBundle,
    Levels,
    LOG,
    Number,
    OverlapStrategy,
    PerImageSliceSpotResults,
    PHYSICAL_COORDINATE_DIMENSION,
    PhysicalCoordinateTypes,
    STARFISH_EXTRAS_KEY,
    TraceBuildingStrategies,
    TransformType,
)



import numpy as np
import pytest
import xarray as xr

from starfish import ImageStack
from starfish.core.image.Filter.call_bases import CallBases
from starfish.types import Axes


def make_multicolor_image():
    '''
        Make a test tile
    '''

    image = np.ones((4, 2, 2), dtype='float32') * 0.25

    x = [0, 0, 1, 1]
    y = [0, 1, 0, 1]

    for i in range(4):
        image[i, x[i], y[i]] = 1

    image[0, 0, 0] = 0.75
    image[2, 1, 1] = 0.9

    image_xr = xr.DataArray(image, dims=(Axes.CH.value, Axes.X.value, Axes.Y.value))

    return image_xr


def make_expected_base_calls(intensity_threshold=0, quality_threshold=0):
    '''
        Get the expected results depending on the test thresholds used.
    '''
    if intensity_threshold == 0 and quality_threshold == 0:
        base_calls = np.array(
            [[
                [0.86602545, 0], [0, 0]],
                [[0, 0.9176629], [0, 0]],
                [[0, 0], [0.9176629, 0]],
                [[0, 0], [0, 0.7188852]]], dtype='float32'
        )

    elif intensity_threshold == 0.8 and quality_threshold == 0:
        base_calls = np.array(
            [[
                [0, 0], [0, 0]],
                [[0, 0.9176629], [0, 0]],
                [[0, 0], [0.9176629, 0]],
                [[0, 0], [0, 0.7188852]]], dtype='float32'
        )

    elif intensity_threshold == 0 and quality_threshold == 0.75:
        base_calls = np.array(
            [[
                [0.86602545, 0], [0, 0]],
                [[0, 0.9176629], [0, 0]],
                [[0, 0], [0.9176629, 0]],
                [[0, 0], [0, 0]]], dtype='float32'
        )

    elif intensity_threshold == 0.8 and quality_threshold == 0.75:
        base_calls = np.array(
            [[
                [0, 0], [0, 0]],
                [[0, 0.9176629], [0, 0]],
                [[0, 0], [0.9176629, 0]],
                [[0, 0], [0, 0]]], dtype='float32'
        )

    bases_xr = xr.DataArray(base_calls, dims=(Axes.CH.value, Axes.X.value, Axes.Y.value))

    return bases_xr


def make_image_stack():
    '''
        Make a test ImageStack

    '''

    # Make the test image
    test = np.ones((1, 4, 1, 2, 2), dtype='float32') * 0.5

    x = [0, 0, 1, 1]
    y = [0, 1, 0, 1]

    for i in range(4):
        test[0, i, 0, x[i], y[i]] = 1
    test[0, 0, 0, 0, 0] = 0.75

    # Make the ImageStack
    test_stack = ImageStack.from_numpy(test)

    return test_stack


def make_expected_image_stack():
    '''
        Make the expected image stack result
    '''

    base_calls = np.array(
        [[[[[0, 0],
          [0, 0]]],
          [[[0, 0.755929],
           [0, 0]]],
          [[[0, 0],
           [0.755929, 0]]],
          [[[0, 0],
           [0, 0.755929]]]]], dtype='float32'
    )

    expected_stack = ImageStack.from_numpy(base_calls)

    return expected_stack


@pytest.mark.parametrize("int_thresh, qual_thresh", [(0.8, 0), (0, 0.75), (0.8, 0.75)])
def test_call_bases(int_thresh: float, qual_thresh: float):

    # Instantiate the filter
    cb = CallBases()

    # Make the image and filter
    image_xr = make_multicolor_image()
    expected_result = make_expected_base_calls(
        intensity_threshold=int_thresh, quality_threshold=qual_thresh
    )
    base_image = cb._call_bases(
        image_xr, intensity_threshold=int_thresh, quality_threshold=qual_thresh
    )

    xr.testing.assert_equal(base_image, expected_result)


def test_image_stack_base_call():

    # Get the test stack and expected result
    test_stack = make_image_stack()
    expected_result = make_expected_image_stack()

    # Filter
    bc = CallBases(intensity_threshold=0.8)
    base_call = bc.run(test_stack, in_place=False)

    xr.testing.assert_equal(base_call.xarray, expected_result.xarray)



import tempfile

import numpy as np

from starfish import data
from starfish.core.image import Filter
from starfish.core.image._registration.LearnTransform.translation import Translation
from starfish.core.image._registration.transforms_list import TransformsList
from starfish.core.types import Axes, TransformType


ISS_SHIFTS = [[-23, 6], [-22, 2], [-22, -3], [-15, -4]]


def test_export_import_transforms_object():
    exp = data.ISS(use_test_data=True)
    stack = exp.fov().get_image('primary')
    reference_stack = exp.fov().get_image('dots')
    translation = Translation(reference_stack=reference_stack, axes=Axes.ROUND)
    # Calculate max_proj accrss CH/Z
    stack = Filter.Reduce((Axes.CH, Axes.ZPLANE)).run(stack)
    transform_list = translation.run(stack)
    _, filename = tempfile.mkstemp()
    # save to tempfile and import
    transform_list.to_json(filename)
    imported = TransformsList.from_json(filename)
    for (_, transform_type, transform), shift in zip(imported.transforms, ISS_SHIFTS):
        # assert that each TransformationObject has the correct translation shift
        assert transform_type == TransformType.SIMILARITY
        assert np.array_equal(transform.translation, shift)



from abc import abstractmethod
from typing import Optional

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.pipeline.algorithmbase import AlgorithmBase


class ApplyTransformAlgorithm(metaclass=AlgorithmBase):

    @abstractmethod
    def run(self, stack, transforms_list, *args) -> Optional[ImageStack]:
        """Performs registration on the stack provided."""
        raise NotImplementedError()



from abc import abstractmethod

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.morphology.binary_mask import BinaryMaskCollection
from starfish.core.pipeline.algorithmbase import AlgorithmBase


class SegmentAlgorithm(metaclass=AlgorithmBase):

    @abstractmethod
    def run(
            self,
            primary_image_stack: ImageStack,
            nuclei_stack: ImageStack,
            *args
    ) -> BinaryMaskCollection:
        """Performs segmentation on the stack provided."""
        raise NotImplementedError()



from typing import Optional

import numpy as np
from scipy.ndimage import label

from starfish.core.morphology.binary_mask import BinaryMaskCollection
from ._base import FilterAlgorithm


class StructuralLabel(FilterAlgorithm):
    """
    Using a BinaryMaskCollection with a single mask, produce a new BinaryMaskCollection with a mask
    representing each feature.  Features are determined using a structuring element, which determine
    what is considered part of a feature.

    Parameters
    ----------
    structure : Optional[np.ndarray]
        See the documentation for :py:func:`scipy.ndimage.label`
    """

    def __init__(self, structure: Optional[np.ndarray] = None) -> None:
        self._structure = structure

    def run(
            self,
            binary_mask_collection: BinaryMaskCollection,
            *args,
            **kwargs
    ) -> BinaryMaskCollection:
        assert len(binary_mask_collection) == 1
        mask = binary_mask_collection.uncropped_mask(0)
        labeled_array, _ = label(mask, self._structure)
        return BinaryMaskCollection.from_label_array_and_ticks(
            labeled_array,
            binary_mask_collection._pixel_ticks,
            binary_mask_collection._physical_ticks,
            binary_mask_collection.log,
        )



import warnings
from functools import partial
from typing import Optional, Tuple

import numpy as np
from trackpy import locate

from starfish.core.image.Filter.util import determine_axes_to_group_by
from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.spots.FindSpots import spot_finding_utils
from starfish.core.types import Axes, PerImageSliceSpotResults, SpotAttributes, SpotFindingResults
from ._base import FindSpotsAlgorithm


class TrackpyLocalMaxPeakFinder(FindSpotsAlgorithm):
    """
    Find spots using a local max peak finding algorithm

    This is a wrapper for :code:`trackpy.locate`, which implements a version of the
    `Crocker-Grier <crocker_grier>`_ algorithm.

    .. _crocker_grier: https://physics.nyu.edu/grierlab/methods3c/

    Parameters
    ----------

    spot_diameter :
        odd integer or tuple of odd integers.
        This may be a single number or a tuple giving the feature’s extent in each dimension,
        useful when the dimensions do not have equal resolution (e.g. confocal microscopy).
        The tuple order is the same as the image shape, conventionally (z, y, x) or (y, x).
        The number(s) must be odd integers. When in doubt, round up.
    min_mass : Optional[float]
        The minimum integrated brightness. This is a crucial parameter for eliminating spurious
        features. Recommended minimum values are 100 for integer images and 1 for float images.
        Defaults to 0 (no filtering).
    max_size : float
        maximum radius-of-gyration of brightness, default None
    separation : Union[float, tuple]
        Minimum separation between features. Default is diameter + 1. May be a tuple, see
        diameter for details.
    percentile : float
        Features must have a peak brighter than pixels in this percentile. This helps eliminate
        spurious peaks. (default = 0)
    noise_size : Union[float, tuple]
        Width of Gaussian blurring kernel, in pixels Default is 1. May be a tuple, see diameter
        for details.
    smoothing_size : Union[float, tuple]
        The size of the sides of the square kernel used in boxcar (rolling average) smoothing,
        in pixels Default is diameter. May be a tuple, making the kernel rectangular.
    threshold : float
        Clip bandpass result below this value. Thresholding is done on the already
        background-subtracted image. By default, 1 for integer images and 1/255 for float
        images.
    measurement_type : str ['max', 'mean']
        name of the function used to calculate the intensity for each identified spot area
        (default = max)
    preprocess : boolean
        Set to False to turn off bandpass prepossessing.
    max_iterations : integer
        Max number of loops to refine the center of mass, (default = 10)
    is_volume : bool
        if True, run the algorithm on 3d volumes of the provided stack (default = False)
    verbose : bool
        If True, report the percentage completed (default = False) during processing

    Notes
    -----
    See also trackpy locate: http://soft-matter.github.io/trackpy/dev/generated/trackpy.locate.html

    """

    def __init__(
            self, spot_diameter, min_mass, max_size, separation, percentile=0,
            noise_size: Tuple[int, int, int] = (1, 1, 1), smoothing_size=None, threshold=None,
            preprocess: bool = False, max_iterations: int = 10, measurement_type: str = 'max',
            is_volume: bool = False, verbose=False, radius_is_gyration: bool = False) -> None:

        self.diameter = spot_diameter
        self.minmass = min_mass
        self.maxsize = max_size
        self.separation = separation
        self.noise_size = noise_size
        self.smoothing_size = smoothing_size
        self.percentile = percentile
        self.threshold = threshold
        self.measurement_function = self._get_measurement_function(measurement_type)
        self.preprocess = preprocess
        self.max_iterations = max_iterations
        self.is_volume = is_volume
        self.verbose = verbose
        self.radius_is_gyration = radius_is_gyration

    def image_to_spots(
            self, data_image: np.ndarray,
    ) -> PerImageSliceSpotResults:
        """

        Parameters
        ----------
        data_image : np.ndarray
            three-dimensional image containing spots to be detected

        Returns
        -------
        PerImageSpotResults :
            includes a SpotAttributes DataFrame of metadata containing the coordinates, intensity
            and radius of each spot, as well as any extra information collected during spot finding.

        """
        data_image = np.asarray(data_image)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', FutureWarning)  # trackpy numpy indexing warning
            warnings.simplefilter('ignore', UserWarning)  # yielded if black images
            attributes = locate(
                data_image,
                diameter=self.diameter,
                minmass=self.minmass,
                maxsize=self.maxsize,
                separation=self.separation,
                noise_size=self.noise_size,
                smoothing_size=self.smoothing_size,
                threshold=self.threshold,
                percentile=self.percentile,
                preprocess=self.preprocess,
                max_iterations=self.max_iterations,
            )

        # when zero spots are detected, 'ep' is missing from the trackpy locate results.
        if attributes.shape[0] == 0:
            attributes['ep'] = []

        # TODO ambrosejcarr: data should always be at least pseudo-3d, this may not be necessary
        # TODO ambrosejcarr: this is where max vs. sum vs. mean would be parametrized.
        # here, total_intensity = sum, intensity = max
        new_colnames = [
            'y', 'x', 'total_intensity', 'radius', 'eccentricity', 'intensity', 'raw_mass', 'ep'
        ]
        if len(data_image.shape) == 3:
            attributes.columns = ['z'] + new_colnames
        else:
            attributes.columns = new_colnames

        attributes['spot_id'] = np.arange(attributes.shape[0])
        return PerImageSliceSpotResults(spot_attrs=SpotAttributes(attributes), extras=None)

    def run(
            self,
            image_stack: ImageStack,
            reference_image: Optional[ImageStack] = None,
            n_processes: Optional[int] = None,
            *args,
    ) -> SpotFindingResults:
        """
        Find spots.

        Parameters
        ----------
        image_stack : ImageStack
            ImageStack where we find the spots in.
        reference_image : Optional[ImageStack]
            (Optional) a reference image. If provided, spots will be found in this image, and then
            the locations that correspond to these spots will be measured across each channel.
        n_processes : Optional[int] = None,
            Number of processes to devote to spot finding.
        """
        spot_finding_method = partial(self.image_to_spots, *args)
        if reference_image:
            data_image = reference_image._squeezed_numpy(*{Axes.ROUND, Axes.CH})
            reference_spots = spot_finding_method(data_image)
            results = spot_finding_utils.measure_intensities_at_spot_locations_across_imagestack(
                image_stack,
                reference_spots,
                measurement_function=self.measurement_function,
                radius_is_gyration=self.radius_is_gyration)
        else:
            spot_attributes_list = image_stack.transform(
                func=spot_finding_method,
                group_by=determine_axes_to_group_by(self.is_volume),
                n_processes=n_processes
            )
            results = SpotFindingResults(imagestack_coords=image_stack.xarray.coords,
                                         log=image_stack.log,
                                         spot_attributes_list=spot_attributes_list)
        return results



import numpy as np
from scipy.ndimage import distance_transform_edt, label
from skimage.feature import peak_local_max
from skimage.morphology import watershed

from starfish.core.morphology.binary_mask import BinaryMaskCollection
from ._base import FilterAlgorithm


class MinDistanceLabel(FilterAlgorithm):
    """
    Using a BinaryMaskCollection with a single mask, produce a new BinaryMaskCollection with a mask
    representing each feature.  Features are determined by obtaining the peak local maximum of a
    matrix consisting of the distance from each position to background.  Watershed is applied to the
    distance matrix to obtain the features.

    Note that due to bugs in skimage, exclude_border should generally be set to False.

    Parameters
    ----------
    minimum_distance_xy : int
        The minimum distance between the peaks along the x or y axis. (default: 1)
    minimum_distance_z : int
        The minimum distance between the peaks along the x or y axis. (default: 1)
    exclude_border : bool
        Exclude the borders for consideration for peaks. (default: False)
    """

    def __init__(
            self,
            minimum_distance_xy: int = 1,
            minimum_distance_z: int = 1,
            exclude_border: bool = False,
    ) -> None:
        self._minimum_distance_xy = minimum_distance_xy
        self._minimum_distance_z = minimum_distance_z
        self._exclude_border = exclude_border

    def run(
            self,
            binary_mask_collection: BinaryMaskCollection,
            *args,
            **kwargs
    ) -> BinaryMaskCollection:
        assert len(binary_mask_collection) == 1
        mask = binary_mask_collection.uncropped_mask(0)

        # calculates the distance of every pixel to the nearest background (0) point
        distance: np.ndarray = distance_transform_edt(mask)

        footprint = np.ones(
            shape=(
                self._minimum_distance_z * 2 + 1,
                self._minimum_distance_xy * 2 + 1,
                self._minimum_distance_xy * 2 + 1,
            ),
            dtype=np.bool,
        )

        # boolean array marking local maxima, excluding any maxima within min_dist
        local_maximum: np.ndarray = peak_local_max(
            distance,
            exclude_border=self._exclude_border,
            indices=False,
            footprint=footprint,
            labels=np.asarray(mask),
        )

        # label the maxima for watershed
        markers, _ = label(local_maximum)

        # run watershed, using the distances in the thresholded image as basins.
        # Uses the original image as a mask, preventing any background pixels from being labeled
        labeled_array: np.ndarray = watershed(-distance, markers, mask=mask)

        return BinaryMaskCollection.from_label_array_and_ticks(
            labeled_array,
            binary_mask_collection._pixel_ticks,
            binary_mask_collection._physical_ticks,
            binary_mask_collection.log,
        )



from . import (
    # configuration management.
    config,
    # image processing methods and objects.
    image,
    # spot detection and manipulation.
    spots,
)
from .core import (
    is_release_tag as __is_release_tag__,
    version as __version__
)
# display images and spots
from .core._display import display
# top-level objects
from .core.codebook.codebook import Codebook
from .core.experiment.experiment import Experiment, FieldOfView
from .core.expression_matrix.expression_matrix import ExpressionMatrix
from .core.imagestack.imagestack import ImageStack
from .core.intensity_table.decoded_intensity_table import DecodedIntensityTable
from .core.intensity_table.intensity_table import IntensityTable
from .core.morphology.binary_mask import BinaryMaskCollection
from .core.morphology.label_image import LabelImage
from .core.segmentation_mask import SegmentationMaskCollection
from .core.util.logging import Log



from typing import Dict, Tuple

from starfish.core.codebook.codebook import Codebook
from starfish.core.intensity_table.decoded_intensity_table import DecodedIntensityTable
from starfish.core.spots.DecodeSpots.trace_builders import build_traces_sequential
from starfish.core.types import Axes, Features, SpotFindingResults
from ._base import DecodeSpotsAlgorithm


class SimpleLookupDecoder(DecodeSpotsAlgorithm):
    """
    Decode spots by assiging the target value of a spot to the corresponding target value of the
    round/ch it was found in. This method only makes sense to use in non mulitplexed sequential
    assays where each r/ch pair only has one target assigned to it.

    Parameters
    ----------
    codebook : Codebook
        Contains codes to decode IntensityTable

    """

    def __init__(self, codebook: Codebook):
        self.codebook = codebook

    def run(self, spots: SpotFindingResults, *args) -> DecodedIntensityTable:
        """
        Decode spots by looking up the associated target value for the round and ch each spot is
        in.

        Parameters
        ----------
        spots: SpotFindingResults
            A Dict of tile indices and their corresponding measured spots

        Returns
        -------
        DecodedIntensityTable :
            IntensityTable decoded and appended with Features.TARGET and values.

        """
        lookup_table: Dict[Tuple, str] = {}
        for target in self.codebook[Features.TARGET]:
            for ch_label in self.codebook[Axes.CH.value]:
                for round_label in self.codebook[Axes.ROUND.value]:
                    if self.codebook.loc[target, round_label, ch_label]:
                        lookup_table[(int(round_label), int(ch_label))] = str(target.values)

        for r_ch_index, results in spots.items():
            target = lookup_table[r_ch_index] if r_ch_index in lookup_table else 'nan'
            results.spot_attrs.data[Features.TARGET] = target
        intensities = build_traces_sequential(spots)
        return DecodedIntensityTable(intensities)



import numpy as np
import pytest

from starfish.core.types import Axes, Coordinates
from ..imagestack import ImageStack


def test_from_numpy_array_raises_error_when_incorrect_dims_passed():
    array = np.ones((2, 2), dtype=np.float32)
    # verify this method works with the correct shape
    image = ImageStack.from_numpy(array.reshape((1, 1, 1, 2, 2)))
    assert isinstance(image, ImageStack)

    with pytest.raises(ValueError):
        ImageStack.from_numpy(array.reshape((1, 1, 2, 2)))
        ImageStack.from_numpy(array.reshape((1, 2, 2)))
        ImageStack.from_numpy(array)
        ImageStack.from_numpy(array.reshape((1, 1, 1, 1, 2, 2)))


def test_from_numpy_array_automatically_handles_float_conversions():
    x = np.zeros((1, 1, 1, 20, 20), dtype=np.uint16)
    stack = ImageStack.from_numpy(x)
    assert stack.xarray.dtype == np.float32


def test_from_numpy_array_coordinates():
    x = np.zeros((1, 1, 4, 20, 20), dtype=np.float32)
    coordinates = {
        Coordinates.X: np.linspace(0, 15, x.shape[-1]),
        Coordinates.Y: np.linspace(0, 20, x.shape[-2]),
        Coordinates.Z: [1, 2, 15, 20],
    }
    stack = ImageStack.from_numpy(x, coordinates=coordinates)

    assert stack.xarray.isel({Axes.X.value: 0}).coords[Coordinates.X.value] == 0
    assert stack.xarray.isel({Axes.X.value: -1}).coords[Coordinates.X.value] == 15
    assert stack.xarray.isel({Axes.Y.value: 0}).coords[Coordinates.Y.value] == 0
    assert stack.xarray.isel({Axes.Y.value: -1}).coords[Coordinates.Y.value] == 20
    assert stack.xarray.isel({Axes.ZPLANE.value: 0}).coords[Coordinates.Z.value] == 1
    assert stack.xarray.isel({Axes.ZPLANE.value: -1}).coords[Coordinates.Z.value] == 20



from abc import abstractmethod
from typing import Callable, Sequence, Tuple

import numpy as np

from starfish.core.imagestack.imagestack import ImageStack
from starfish.core.intensity_table.intensity_table import IntensityTable
from starfish.core.pipeline.algorithmbase import AlgorithmBase
from starfish.core.types import Number
from .combine_adjacent_features import ConnectedComponentDecodingResult


class DetectPixelsAlgorithm(metaclass=AlgorithmBase):

    @abstractmethod
    def run(
            self,
            primary_image: ImageStack,
            *args,
    ) -> Tuple[IntensityTable, ConnectedComponentDecodingResult]:
        """Finds spots in an ImageStack"""
        raise NotImplementedError()

    @staticmethod
    def _get_measurement_function(measurement_type: str) -> Callable[[Sequence], Number]:
        try:
            measurement_function = getattr(np, measurement_type)
        except AttributeError:
            raise ValueError(
                f'measurement_type must be a numpy reduce function such as "max" or "mean". '
                f'{measurement_type} not found.')
        return measurement_function



from functools import partial
from typing import Callable, Optional, Tuple, Union

import numpy as np
import xarray as xr
from scipy.ndimage.filters import uniform_filter

from starfish.core.imagestack.imagestack import _reconcile_clip_and_level, ImageStack
from starfish.core.types import Clip, Levels, Number
from ._base import FilterAlgorithm
from .util import (
    determine_axes_to_group_by, validate_and_broadcast_kernel_size
)


class MeanHighPass(FilterAlgorithm):
    """
    The mean high pass filter reduces low spatial frequency features by subtracting a
    mean filtered image from the original image. The mean filter smooths an image by replacing
    each pixel's value with an average of the pixel values of the surrounding neighborhood.

    The mean filter is also known as a uniform or box filter. It can also be considered as a fast
    approximation to a GaussianHighPass filter.

    This is a pass through for the scipy.ndimage.filters.uniform_filter:
    https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.ndimage.uniform_filter.html

    Parameters
    ----------
    size : Union[Number, Tuple[Number]]
        width of the kernel
    is_volume : bool
        If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
        independently.
    clip_method : Optional[Union[str, :py:class:`~starfish.types.Clip`]]
        Deprecated method to control the way that data are scaled to retain skimage dtype
        requirements that float data fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Clip.CLIP: data above 1 are set to 1.  This has been replaced with
          level_method=Levels.CLIP.
        - Clip.SCALE_BY_IMAGE: when any data in the entire ImageStack is greater than 1, the entire
          ImageStack is scaled by the maximum value in the ImageStack.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_IMAGE.
        - Clip.SCALE_BY_CHUNK: when any data in any slice is greater than 1, each slice is scaled by
          the maximum value found in that slice.  The slice shapes are determined by the
          ``group_by`` parameters.  This has been replaced with
          level_method=Levels.SCALE_SATURATED_BY_CHUNK.
    level_method : :py:class:`~starfish.types.Levels`
        Controls the way that data are scaled to retain skimage dtype requirements that float data
        fall in [0, 1].  In all modes, data below 0 are set to 0.

        - Levels.CLIP (default): data above 1 are set to 1.
        - Levels.SCALE_SATURATED_BY_IMAGE: when any data in the entire ImageStack is greater
          than 1, the entire ImageStack is scaled by the maximum value in the ImageStack.
        - Levels.SCALE_SATURATED_BY_CHUNK: when any data in any slice is greater than 1, each
          slice is scaled by the maximum value found in that slice.  The slice shapes are
          determined by the ``group_by`` parameters.
        - Levels.SCALE_BY_IMAGE: scale the entire ImageStack by the maximum value in the
          ImageStack.
        - Levels.SCALE_BY_CHUNK: scale each slice by the maximum value found in that slice.  The
          slice shapes are determined by the ``group_by`` parameters.
    """

    def __init__(
            self,
            size: Union[Number, Tuple[Number]],
            is_volume: bool = False,
            clip_method: Optional[Union[str, Clip]] = None,
            level_method: Optional[Levels] = None
    ) -> None:

        self.size = validate_and_broadcast_kernel_size(size, is_volume)
        self.is_volume = is_volume
        self.level_method = _reconcile_clip_and_level(clip_method, level_method)

    _DEFAULT_TESTING_PARAMETERS = {"size": 1}

    @staticmethod
    def _high_pass(
        image: xr.DataArray, size: Number, rescale: bool = False
    ) -> xr.DataArray:
        """
        Applies a mean high pass filter to an image

        Parameters
        ----------
        image : xr.DataArray
            2-d or 3-d image data
        size : Union[Number, Tuple[Number]]
            width of the kernel
        rescale : bool
            If true scales data by max value, if false clips max values to one

        Returns
        -------
        xr.DataArray:
            Filtered image, same shape as input
        """

        blurred: np.ndarray = uniform_filter(image, size)

        filtered: np.ndarray = image - blurred

        return filtered

    def run(
            self,
            stack: ImageStack,
            in_place: bool = False,
            verbose: bool = False,
            n_processes: Optional[int] = None,
            *args,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to applying the filter. If None, defaults to
            the result of os.cpu_count(). (default None)

        Returns
        -------
        ImageStack :
            If in-place is False, return the results of filter as a new stack. Otherwise return the
            original stack.

        """
        group_by = determine_axes_to_group_by(self.is_volume)
        high_pass: Callable = partial(self._high_pass, size=self.size)
        result = stack.apply(
            high_pass,
            group_by=group_by, verbose=verbose, in_place=in_place, n_processes=n_processes,
            level_method=self.level_method
        )
        return result



#!/usr/bin/env python

import os
from pathlib import Path
import setuptools
import versioneer

install_requires = [
    line.rstrip() for line in open(Path(__file__).parent / "REQUIREMENTS.txt")
]

with open(Path("starfish") / "core" / "_display.py") as f:
    for line in f.readlines():
        if line.startswith("NAPARI_VERSION"):
            napari_version = line.split('"')[1]
            break

setuptools.setup(
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    extras_require={
        'napari': [f"napari>={napari_version}"],
    },
    entry_points={
        'console_scripts': [
            "starfish=starfish.core.starfish:starfish",
        ]
    },
    include_package_data=True,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



import numpy as np
import pytest

from starfish.core.types import Axes, Coordinates
from .factories import binary_arrays_2d, binary_arrays_3d
from ..binary_mask import BinaryMaskCollection


def test_2d():
    """Simple case of BinaryMaskCollection.from_binary_arrays_and_ticks with 2D data.  Pixel ticks
    are inferred."""
    binary_arrays, physical_ticks = binary_arrays_2d()

    binary_mask_collection = BinaryMaskCollection.from_binary_arrays_and_ticks(
        binary_arrays,
        None,
        physical_ticks,
        None
    )
    assert len(binary_mask_collection) == 2

    region_0, region_1 = binary_mask_collection.masks()
    assert region_0.name == '0'
    assert region_1.name == '1'

    assert np.array_equal(region_0, np.ones((1, 6), dtype=np.bool))
    temp = np.ones((2, 3), dtype=np.bool)
    temp[-1, -1] = False
    assert np.array_equal(region_1, temp)

    assert np.array_equal(region_0[Axes.Y.value], [0])
    assert np.array_equal(region_0[Axes.X.value], [0, 1, 2, 3, 4, 5])

    assert np.array_equal(region_1[Axes.Y.value], [3, 4])
    assert np.array_equal(region_1[Axes.X.value], [3, 4, 5])

    assert np.array_equal(region_0[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][0:1])
    assert np.array_equal(region_0[Coordinates.X.value],
                          physical_ticks[Coordinates.X][0:6])

    assert np.array_equal(region_1[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][3:6])
    assert np.array_equal(region_1[Coordinates.X.value],
                          physical_ticks[Coordinates.X][3:6])


def test_3d():
    """Simple case of BinaryMaskCollection.from_binary_arrays_and_ticks with 3D data.  Pixel ticks
    are inferred."""
    binary_arrays, physical_ticks = binary_arrays_3d()

    binary_mask_collection = BinaryMaskCollection.from_binary_arrays_and_ticks(
        binary_arrays,
        None,
        physical_ticks,
        None
    )
    assert len(binary_mask_collection) == 2

    region_0, region_1 = binary_mask_collection.masks()
    assert region_0.name == '0'
    assert region_1.name == '1'

    assert np.array_equal(region_0, np.ones((1, 1, 6), dtype=np.bool))
    temp = np.ones((2, 2, 3), dtype=np.bool)
    temp[-1, -1, -1] = False
    assert np.array_equal(region_1, temp)

    assert np.array_equal(region_0[Axes.ZPLANE.value], [0])
    assert np.array_equal(region_0[Axes.Y.value], [0])
    assert np.array_equal(region_0[Axes.X.value], [0, 1, 2, 3, 4, 5])

    assert np.array_equal(region_1[Axes.ZPLANE.value], [0, 1])
    assert np.array_equal(region_1[Axes.Y.value], [3, 4])
    assert np.array_equal(region_1[Axes.X.value], [3, 4, 5])

    assert np.array_equal(region_0[Coordinates.Z.value],
                          physical_ticks[Coordinates.Z][0:1])
    assert np.array_equal(region_0[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][0:1])
    assert np.array_equal(region_0[Coordinates.X.value],
                          physical_ticks[Coordinates.X][0:6])

    assert np.array_equal(region_1[Coordinates.Z.value],
                          physical_ticks[Coordinates.Z][0:2])
    assert np.array_equal(region_1[Coordinates.Y.value],
                          physical_ticks[Coordinates.Y][3:5])
    assert np.array_equal(region_1[Coordinates.X.value],
                          physical_ticks[Coordinates.X][3:6])


def test_no_mask():
    """Simple case of BinaryMaskCollection.from_binary_arrays_and_ticks with no masks.  Pixel ticks
    are inferred."""
    _, physical_ticks = binary_arrays_2d()

    binary_mask_collection = BinaryMaskCollection.from_binary_arrays_and_ticks(
        [],
        None,
        physical_ticks,
        None
    )
    assert len(binary_mask_collection) == 0

    assert np.array_equal(binary_mask_collection._pixel_ticks[Axes.X], np.arange(0, 6))
    assert np.array_equal(binary_mask_collection._pixel_ticks[Axes.Y], np.arange(0, 5))
    assert np.array_equal(
        binary_mask_collection._physical_ticks[Coordinates.X],
        physical_ticks[Coordinates.X])
    assert np.array_equal(
        binary_mask_collection._physical_ticks[Coordinates.Y],
        physical_ticks[Coordinates.Y])


def test_empty_mask():
    """Simple case of BinaryMaskCollection.from_binary_arrays_and_ticks with no masks.  Pixel ticks
    are inferred."""
    binary_arrays = [
        np.zeros((5, 6), dtype=np.bool),
    ]
    _, physical_ticks = binary_arrays_2d()

    binary_mask_collection = BinaryMaskCollection.from_binary_arrays_and_ticks(
        binary_arrays,
        None,
        physical_ticks,
        None
    )
    assert len(binary_mask_collection) == 1

    assert binary_mask_collection[0].shape == (0, 0)

    assert np.array_equal(binary_mask_collection._pixel_ticks[Axes.X], np.arange(0, 6))
    assert np.array_equal(binary_mask_collection._pixel_ticks[Axes.Y], np.arange(0, 5))
    assert np.array_equal(
        binary_mask_collection._physical_ticks[Coordinates.X],
        physical_ticks[Coordinates.X])
    assert np.array_equal(
        binary_mask_collection._physical_ticks[Coordinates.Y],
        physical_ticks[Coordinates.Y])


def test_mismatched_binary_array_sizes():
    """Simple case of BinaryMaskCollection.from_binary_arrays_and_ticks with 2D data.  Not all
    arrays are sized identically."""
    binary_arrays = [
        np.zeros((3, 6), dtype=np.bool),
        np.zeros((5, 6), dtype=np.bool),
    ]
    binary_arrays[0][0] = True
    binary_arrays[1][3:5, 3:6] = True
    binary_arrays[1][-1, -1] = False

    _, physical_ticks = binary_arrays_2d()

    with pytest.raises(ValueError):
        BinaryMaskCollection.from_binary_arrays_and_ticks(
            binary_arrays,
            None,
            physical_ticks,
            None
        )


def test_mismatched_binary_array_types():
    """Simple case of BinaryMaskCollection.from_binary_arrays_and_ticks with 2D data.  Not all
    arrays are of the correct type."""
    binary_arrays = [
        np.zeros((3, 6), dtype=np.bool),
        np.zeros((3, 6), dtype=np.int),
    ]
    binary_arrays[0][0] = True
    binary_arrays[1][3:5, 3:6] = True
    binary_arrays[1][-1, -1] = False

    _, physical_ticks = binary_arrays_2d()

    with pytest.raises(TypeError):
        BinaryMaskCollection.from_binary_arrays_and_ticks(
            binary_arrays,
            None,
            physical_ticks,
            None
        )


def test_incorrectly_sized_pixel_ticks():
    """BinaryMaskCollection.from_binary_arrays_and_ticks with 2D data.  Pixel ticks are incorrectly
    sized."""
    binary_arrays, physical_ticks = binary_arrays_2d()

    pixel_ticks = {
        Axes.Y: [0, 1, 2, 3],
        Axes.X: [0, 1, 2, 3, 4, 5],
    }

    with pytest.raises(ValueError):
        BinaryMaskCollection.from_binary_arrays_and_ticks(
            binary_arrays,
            pixel_ticks,
            physical_ticks,
            None
        )


def test_missing_physical_ticks():
    """BinaryMaskCollection.from_binary_arrays_and_ticks with some physical ticks missing."""
    binary_arrays = [
        np.zeros((2, 5, 6), dtype=np.bool),
        np.zeros((2, 5, 6), dtype=np.bool),
    ]

    physical_ticks = {
        Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
        Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12, 13.2]}

    # Z physical ticks missing
    with pytest.raises(ValueError):
        BinaryMaskCollection.from_binary_arrays_and_ticks(
            binary_arrays,
            None,
            physical_ticks,
            None
        )


def test_incorrectly_sized_physical_ticks():
    """BinaryMaskCollection.from_label_array_and_ticks with some physical ticks incorrectly
    sized."""
    binary_2d_arrays = [
        np.zeros((5, 6), dtype=np.bool),
        np.zeros((5, 6), dtype=np.bool),
    ]

    physical_ticks_2d = {Coordinates.Y: [1.2, 2.4, 3.6, 4.8, 6.0],
                         Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12]}

    with pytest.raises(ValueError):
        BinaryMaskCollection.from_binary_arrays_and_ticks(
            binary_2d_arrays,
            None,
            physical_ticks_2d,
            None
        )

    binary_3d_arrays = [
        np.zeros((2, 5, 6), dtype=np.bool),
        np.zeros((2, 5, 6), dtype=np.bool),
    ]

    physical_ticks_3d = {
        Coordinates.Z: [0.0, 1.0],
        Coordinates.Y: [1.2, 2.4, 3.6, 4.8],
        Coordinates.X: [7.2, 8.4, 9.6, 10.8, 12, 13.2]
    }

    with pytest.raises(ValueError):
        BinaryMaskCollection.from_binary_arrays_and_ticks(
            binary_3d_arrays,
            None,
            physical_ticks_3d,
            None
        )



import numpy as np
import pytest

from starfish import ImageStack
from starfish.types import Number
from ..threshold import ThresholdBinarize


@pytest.mark.parametrize(["threshold"], [[threshold] for threshold in np.linspace(0, 1, 3)])
def test_binarize(threshold: Number, num_rounds=1, num_chs=1, num_zplanes=4, ysize=5, xsize=6):
    data = np.linspace(0, 1, num_rounds * num_chs * num_zplanes * ysize * xsize, dtype=np.float32)
    data = data.reshape((num_rounds, num_chs, num_zplanes, ysize, xsize))

    imagestack = ImageStack.from_numpy(data)
    binarizer = ThresholdBinarize(threshold)
    binary_mask_collection = binarizer.run(imagestack)

    assert len(binary_mask_collection) == 1
    mask = binary_mask_collection.uncropped_mask(0)

    expected_value = data[0, 0] >= threshold

    assert np.array_equal(mask, expected_value)


@pytest.mark.parametrize(
    ["num_rounds", "num_chs"],
    [
        [1, 2],
        [2, 1],
        [2, 2],
    ])
def test_binarize_non_3d(num_rounds, num_chs, num_zplanes=4, ysize=5, xsize=6):
    data = np.linspace(0, 1, num_rounds * num_chs * num_zplanes * ysize * xsize, dtype=np.float32)
    data = data.reshape((num_rounds, num_chs, num_zplanes, ysize, xsize))

    imagestack = ImageStack.from_numpy(data)
    binarizer = ThresholdBinarize(0.0)

    with pytest.raises(ValueError):
        binarizer.run(imagestack)
